from dagster._core.pipes.client import PipesClientCompletedInvocation
from dagster._core.pipes.context import PipesExecutionResult, PipesSession
from dagster._core.pipes.utils import open_pipes_session

__all__ = [
    "PipesClientCompletedInvocation",
    "PipesExecutionResult",
    "PipesSession",
    "open_pipes_session",
]



import re


def clean_asset_name(name: str) -> str:
    """Cleans an input to be a valid Dagster asset name by replacing invalid characters with underscores."""
    return re.sub(r"[^A-Za-z0-9_]+", "_", name)



from collections.abc import Iterator, Mapping, Sequence
from typing import AbstractSet, Any, Optional  # noqa: UP035

import dagster._check as check
from dagster._annotations import deprecated, public
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.data_version import DataProvenance, DataVersion
from dagster._core.definitions.dependency import Node, NodeHandle
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey, UserEvent
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.definitions.step_launcher import StepLauncher
from dagster._core.definitions.time_window_partitions import TimeWindow
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.events import DagsterEvent
from dagster._core.execution.context.op_execution_context import OpExecutionContext
from dagster._core.execution.context.system import StepExecutionContext
from dagster._core.instance import DagsterInstance
from dagster._core.log_manager import DagsterLogManager
from dagster._core.storage.dagster_run import DagsterRun
from dagster._utils.forked_pdb import ForkedPdb


def _copy_docs_from_op_execution_context(obj):
    setattr(obj, "__doc__", getattr(OpExecutionContext, obj.__name__).__doc__)
    return obj


ALTERNATE_METHODS = {
    "run_id": "run.run_id",
    "dagster_run": "run",
    "run_config": "run.run_config",
    "run_tags": "run.tags",
    "get_op_execution_context": "op_execution_context",
    "asset_partition_key_for_output": "partition_key",
    "asset_partitions_time_window_for_output": "partition_time_window",
    "asset_partition_key_range_for_output": "partition_key_range",
    "asset_partitions_def_for_output": "assets_def.partitions_def",
    "asset_partition_keys_for_output": "partition_keys",
}

ALTERNATE_EXPRESSIONS = {
    "has_tag": "key in context.run.tags",
    "get_tag": "context.run.tags.get(key)",
}

USE_OP_CONTEXT = [
    "op_config",
    "node_handle",
    "op_handle",
    "op",
    "get_mapping_key",
    "selected_output_names",
]


def _get_deprecation_kwargs(attr: str) -> Mapping[str, Any]:
    deprecation_kwargs = {"breaking_version": "a future release"}
    deprecation_kwargs["subject"] = f"AssetExecutionContext.{attr}"

    if attr in ALTERNATE_METHODS:
        deprecation_kwargs["additional_warn_text"] = (
            f"You have called the deprecated method {attr} on AssetExecutionContext. Use"
            f" context.{ALTERNATE_METHODS[attr]} instead."
        )
    elif attr in ALTERNATE_EXPRESSIONS:
        deprecation_kwargs["additional_warn_text"] = (
            f"You have called the deprecated method {attr} on AssetExecutionContext. Use"
            f" {ALTERNATE_EXPRESSIONS[attr]} instead."
        )
    elif attr in USE_OP_CONTEXT:
        deprecation_kwargs["additional_warn_text"] = (
            f"You have called the deprecated method {attr} on AssetExecutionContext. Use"
            f" context.op_execution_context.{attr} instead."
        )

    return deprecation_kwargs


class AssetExecutionContext:
    def __init__(self, op_execution_context: OpExecutionContext) -> None:
        self._op_execution_context = check.inst_param(
            op_execution_context, "op_execution_context", OpExecutionContext
        )
        self._step_execution_context = self._op_execution_context._step_execution_context  # noqa: SLF001

    @staticmethod
    def get() -> "AssetExecutionContext":
        from dagster._core.execution.context.asset_check_execution_context import (
            AssetCheckExecutionContext,
        )
        from dagster._core.execution.context.compute import current_execution_context

        ctx = current_execution_context.get()
        if ctx is None:
            raise DagsterInvariantViolationError("No current AssetExecutionContext in scope.")
        if isinstance(ctx, AssetCheckExecutionContext):
            raise DagsterInvariantViolationError(
                "Can't use AssetExecutionContext.get() in the context of an "
                "AssetCheckExecutionContext. Use AssetCheckExecutionContext.get() instead."
            )
        return ctx

    @property
    def op_execution_context(self) -> OpExecutionContext:
        return self._op_execution_context

    ####### Top-level properties/methods on AssetExecutionContext

    @public
    @property
    def log(self) -> DagsterLogManager:
        """The log manager available in the execution context. Logs will be viewable in the Dagster UI.
        Returns: DagsterLogManager.

        Example:
            .. code-block:: python

                @asset
                def logger(context):
                    context.log.info("Info level message")
        """
        return self.op_execution_context.log

    @public
    @property
    def pdb(self) -> ForkedPdb:
        """Gives access to pdb debugging from within the asset. Materializing the asset via the
        Dagster UI or CLI will enter the pdb debugging context in the process used to launch the UI or
        run the CLI.

        Returns: dagster.utils.forked_pdb.ForkedPdb

        Example:
            .. code-block:: python

                @asset
                def debug(context):
                    context.pdb.set_trace()
        """
        return self.op_execution_context.pdb

    @property
    def run(self) -> DagsterRun:
        """The DagsterRun object corresponding to the execution. Information like run_id,
        run configuration, and the assets selected for materialization can be found on the DagsterRun.
        """
        return self.op_execution_context.run

    @public
    @property
    def job_def(self) -> JobDefinition:
        """The definition for the currently executing job. Information like the job name, and job tags
        can be found on the JobDefinition.
        Returns: JobDefinition.
        """
        return self.op_execution_context.job_def

    @property
    def repository_def(self) -> RepositoryDefinition:
        """RepositoryDefinition: The Dagster repository for the currently executing job."""
        return self.op_execution_context.repository_def

    ######## Deprecated methods

    @deprecated(**_get_deprecation_kwargs("dagster_run"))
    @property
    @_copy_docs_from_op_execution_context
    def dagster_run(self) -> DagsterRun:
        return self.op_execution_context.dagster_run

    @deprecated(**_get_deprecation_kwargs("run_id"))
    @property
    @_copy_docs_from_op_execution_context
    def run_id(self) -> str:
        return self.op_execution_context.run_id

    @deprecated(**_get_deprecation_kwargs("run_config"))
    @property
    @_copy_docs_from_op_execution_context
    def run_config(self) -> Mapping[str, object]:
        return self.op_execution_context.run_config

    @deprecated(**_get_deprecation_kwargs("run_tags"))
    @property
    @_copy_docs_from_op_execution_context
    def run_tags(self) -> Mapping[str, str]:
        return self.op_execution_context.run_tags

    @deprecated(**_get_deprecation_kwargs("has_tag"))
    @public
    @_copy_docs_from_op_execution_context
    def has_tag(self, key: str) -> bool:
        return self.op_execution_context.has_tag(key)

    @deprecated(**_get_deprecation_kwargs("get_tag"))
    @public
    @_copy_docs_from_op_execution_context
    def get_tag(self, key: str) -> Optional[str]:
        return self.op_execution_context.get_tag(key)

    @deprecated(**_get_deprecation_kwargs("get_op_execution_context"))
    def get_op_execution_context(self) -> "OpExecutionContext":
        return self.op_execution_context

    @deprecated(**_get_deprecation_kwargs("asset_partition_key_for_output"))
    @public
    @_copy_docs_from_op_execution_context
    def asset_partition_key_for_output(self, output_name: str = "result") -> str:
        return self.op_execution_context.asset_partition_key_for_output(output_name=output_name)

    @deprecated(**_get_deprecation_kwargs("asset_partitions_time_window_for_output"))
    @public
    @_copy_docs_from_op_execution_context
    def asset_partitions_time_window_for_output(self, output_name: str = "result") -> TimeWindow:
        return self.op_execution_context.asset_partitions_time_window_for_output(output_name)

    @deprecated(**_get_deprecation_kwargs("asset_partition_key_range_for_output"))
    @public
    @_copy_docs_from_op_execution_context
    def asset_partition_key_range_for_output(
        self, output_name: str = "result"
    ) -> PartitionKeyRange:
        return self.op_execution_context.asset_partition_key_range_for_output(output_name)

    @deprecated(**_get_deprecation_kwargs("asset_partitions_def_for_output"))
    @public
    @_copy_docs_from_op_execution_context
    def asset_partitions_def_for_output(self, output_name: str = "result") -> PartitionsDefinition:
        return self.op_execution_context.asset_partitions_def_for_output(output_name=output_name)

    @deprecated(**_get_deprecation_kwargs("asset_partition_keys_for_output"))
    @public
    @_copy_docs_from_op_execution_context
    def asset_partition_keys_for_output(self, output_name: str = "result") -> Sequence[str]:
        return self.op_execution_context.asset_partition_keys_for_output(output_name=output_name)

    @deprecated(**_get_deprecation_kwargs("op_config"))
    @public
    @property
    @_copy_docs_from_op_execution_context
    def op_config(self) -> Any:
        return self.op_execution_context.op_config

    @deprecated(**_get_deprecation_kwargs("node_handle"))
    @property
    @_copy_docs_from_op_execution_context
    def node_handle(self) -> NodeHandle:
        return self.op_execution_context.node_handle

    @deprecated(**_get_deprecation_kwargs("op_handle"))
    @property
    @_copy_docs_from_op_execution_context
    def op_handle(self) -> NodeHandle:
        return self.op_execution_context.op_handle

    @deprecated(**_get_deprecation_kwargs("op"))
    @property
    @_copy_docs_from_op_execution_context
    def op(self) -> Node:
        return self.op_execution_context.op

    @deprecated(**_get_deprecation_kwargs("get_mapping_key"))
    @public
    @_copy_docs_from_op_execution_context
    def get_mapping_key(self) -> Optional[str]:
        return self.op_execution_context.get_mapping_key()

    @deprecated(**_get_deprecation_kwargs("selected_output_names"))
    @public
    @property
    @_copy_docs_from_op_execution_context
    def selected_output_names(self) -> AbstractSet[str]:
        return self.op_execution_context.selected_output_names

    ########## pass-through to op context

    #### op related

    @property
    @_copy_docs_from_op_execution_context
    def retry_number(self):
        return self.op_execution_context.retry_number

    @_copy_docs_from_op_execution_context
    def describe_op(self) -> str:
        return self.op_execution_context.describe_op()

    @public
    @property
    @_copy_docs_from_op_execution_context
    def op_def(self) -> OpDefinition:
        return self.op_execution_context.op_def

    #### job related

    @public
    @property
    @_copy_docs_from_op_execution_context
    def job_name(self) -> str:
        return self.op_execution_context.job_name

    #### asset related

    @public
    @property
    @_copy_docs_from_op_execution_context
    def asset_key(self) -> AssetKey:
        return self.op_execution_context.asset_key

    @public
    @property
    @_copy_docs_from_op_execution_context
    def has_assets_def(self) -> bool:
        return self.op_execution_context.has_assets_def

    @public
    @property
    @_copy_docs_from_op_execution_context
    def assets_def(self) -> AssetsDefinition:
        return self.op_execution_context.assets_def

    @public
    @_copy_docs_from_op_execution_context
    def asset_key_for_output(self, output_name: str = "result") -> AssetKey:
        return self.op_execution_context.asset_key_for_output(output_name=output_name)

    @public
    @_copy_docs_from_op_execution_context
    def output_for_asset_key(self, asset_key: AssetKey) -> str:
        return self.op_execution_context.output_for_asset_key(asset_key=asset_key)

    @public
    @_copy_docs_from_op_execution_context
    def asset_key_for_input(self, input_name: str) -> AssetKey:
        return self.op_execution_context.asset_key_for_input(input_name=input_name)

    @public
    @property
    @_copy_docs_from_op_execution_context
    def selected_asset_keys(self) -> AbstractSet[AssetKey]:
        return self.op_execution_context.selected_asset_keys

    #### execution related

    @public
    @property
    @_copy_docs_from_op_execution_context
    def instance(self) -> DagsterInstance:
        return self.op_execution_context.instance

    @property
    @_copy_docs_from_op_execution_context
    def step_launcher(self) -> Optional[StepLauncher]:
        return self.op_execution_context.step_launcher

    @_copy_docs_from_op_execution_context
    def get_step_execution_context(self) -> StepExecutionContext:
        return self.op_execution_context.get_step_execution_context()

    #### partition_related

    @public
    @property
    @_copy_docs_from_op_execution_context
    def has_partition_key(self) -> bool:
        return self.op_execution_context.has_partition_key

    @public
    @property
    @_copy_docs_from_op_execution_context
    def partition_key(self) -> str:
        return self.op_execution_context.partition_key

    @public
    @property
    @_copy_docs_from_op_execution_context
    def partition_keys(self) -> Sequence[str]:
        return self.op_execution_context.partition_keys

    @public
    @property
    @_copy_docs_from_op_execution_context
    def has_partition_key_range(self) -> bool:
        return self.op_execution_context.has_partition_key_range

    @deprecated(breaking_version="2.0", additional_warn_text="Use `partition_key_range` instead.")
    @public
    @property
    @_copy_docs_from_op_execution_context
    def asset_partition_key_range(self) -> PartitionKeyRange:
        return self.op_execution_context.asset_partition_key_range

    @public
    @property
    @_copy_docs_from_op_execution_context
    def partition_key_range(self) -> PartitionKeyRange:
        return self.op_execution_context.partition_key_range

    @public
    @property
    @_copy_docs_from_op_execution_context
    def partition_time_window(self) -> TimeWindow:
        return self.op_execution_context.partition_time_window

    @public
    @_copy_docs_from_op_execution_context
    def asset_partition_key_range_for_input(self, input_name: str) -> PartitionKeyRange:
        return self.op_execution_context.asset_partition_key_range_for_input(input_name)

    @public
    @_copy_docs_from_op_execution_context
    def asset_partition_key_for_input(self, input_name: str) -> str:
        return self.op_execution_context.asset_partition_key_for_input(input_name)

    @public
    @_copy_docs_from_op_execution_context
    def asset_partitions_def_for_input(self, input_name: str) -> PartitionsDefinition:
        return self.op_execution_context.asset_partitions_def_for_input(input_name=input_name)

    @public
    @_copy_docs_from_op_execution_context
    def asset_partition_keys_for_input(self, input_name: str) -> Sequence[str]:
        return self.op_execution_context.asset_partition_keys_for_input(input_name=input_name)

    @public
    @_copy_docs_from_op_execution_context
    def asset_partitions_time_window_for_input(self, input_name: str = "result") -> TimeWindow:
        return self.op_execution_context.asset_partitions_time_window_for_input(input_name)

    #### Event log related

    @_copy_docs_from_op_execution_context
    def has_events(self) -> bool:
        return self.op_execution_context.has_events()

    @_copy_docs_from_op_execution_context
    def consume_events(self) -> Iterator[DagsterEvent]:
        yield from self.op_execution_context.consume_events()

    @public
    @_copy_docs_from_op_execution_context
    def log_event(self, event: UserEvent) -> None:
        return self.op_execution_context.log_event(event)

    #### metadata related

    @public
    @_copy_docs_from_op_execution_context
    def add_output_metadata(
        self,
        metadata: Mapping[str, Any],
        output_name: Optional[str] = None,
        mapping_key: Optional[str] = None,
    ) -> None:
        return self.op_execution_context.add_output_metadata(
            metadata=metadata,
            output_name=output_name,
            mapping_key=mapping_key,
        )

    @public
    def add_asset_metadata(
        self,
        metadata: Mapping[str, Any],
        asset_key: Optional[CoercibleToAssetKey] = None,
        partition_key: Optional[str] = None,
    ) -> None:
        """Add metadata to an asset materialization event. This metadata will be
        available in the Dagster UI.

        Args:
            metadata (Mapping[str, Any]): The metadata to add to the asset
                materialization event.
            asset_key (Optional[CoercibleToAssetKey]): The asset key to add metadata to.
                Does not need to be provided if only one asset is currently being
                materialized.
            partition_key (Optional[str]): The partition key to add metadata to, if
                applicable. Should not be provided on non-partitioned assets. If not
                provided on a partitioned asset, the metadata will be added to all
                partitions of the asset currently being materialized.

        Examples:
            Adding metadata to the asset materialization event for a single asset:

            .. code-block:: python

                import dagster as dg

                @dg.asset
                def my_asset(context):
                    # Add metadata
                    context.add_asset_metadata({"key": "value"})

            Adding metadata to the asset materialization event for a particular partition of a partitioned asset:

            .. code-block:: python

                import dagster as dg

                @dg.asset(partitions_def=dg.StaticPartitionsDefinition(["a", "b"]))
                def my_asset(context):
                    # Adds metadata to all partitions currently being materialized, since no
                    # partition is specified.
                    context.add_asset_metadata({"key": "value"})

                    for partition_key in context.partition_keys:
                        # Add metadata only to the event for partition "a"
                        if partition_key == "a":
                            context.add_asset_metadata({"key": "value"}, partition_key=partition_key)

            Adding metadata to the asset materialization event for a particular asset in a multi-asset.

            .. code-block:: python

                import dagster as dg

                @dg.multi_asset(specs=[dg.AssetSpec("asset1"), dg.AssetSpec("asset2")])
                def my_multi_asset(context):
                    # Add metadata to the materialization event for "asset1"
                    context.add_asset_metadata({"key": "value"}, asset_key="asset1")

                    # THIS line will fail since asset key is not specified:
                    context.add_asset_metadata({"key": "value"})

        """
        self._step_execution_context.add_asset_metadata(
            metadata=metadata,
            asset_key=asset_key,
            partition_key=partition_key,
        )

    @_copy_docs_from_op_execution_context
    def get_output_metadata(
        self,
        output_name: str,
        mapping_key: Optional[str] = None,
    ) -> Optional[Mapping[str, Any]]:
        return self.op_execution_context.get_output_metadata(
            output_name=output_name,
            mapping_key=mapping_key,
        )

    #### asset check related

    @public
    @property
    @_copy_docs_from_op_execution_context
    def selected_asset_check_keys(self) -> AbstractSet[AssetCheckKey]:
        return self.op_execution_context.selected_asset_check_keys

    #### data lineage related

    @public
    @_copy_docs_from_op_execution_context
    def get_asset_provenance(self, asset_key: AssetKey) -> Optional[DataProvenance]:
        return self.op_execution_context.get_asset_provenance(asset_key=asset_key)

    @_copy_docs_from_op_execution_context
    def set_data_version(self, asset_key: AssetKey, data_version: DataVersion) -> None:
        return self.op_execution_context.set_data_version(
            asset_key=asset_key, data_version=data_version
        )

    # misc

    @public
    @property
    @_copy_docs_from_op_execution_context
    def resources(self) -> Any:
        return self.op_execution_context.resources

    @property
    @_copy_docs_from_op_execution_context
    def is_subset(self):
        return self.op_execution_context.is_subset

    # In this mode no conversion is done on returned values and missing but expected outputs are not
    # allowed.
    @property
    @_copy_docs_from_op_execution_context
    def requires_typed_event_stream(self) -> bool:
        return self.op_execution_context.requires_typed_event_stream

    @property
    @_copy_docs_from_op_execution_context
    def typed_event_stream_error_message(self) -> Optional[str]:
        return self.op_execution_context.typed_event_stream_error_message

    @_copy_docs_from_op_execution_context
    def set_requires_typed_event_stream(self, *, error_message: Optional[str] = None) -> None:
        self.op_execution_context.set_requires_typed_event_stream(error_message=error_message)



import functools
from collections.abc import Iterator, Mapping, Sequence
from enum import Enum
from hashlib import sha256
from typing import TYPE_CHECKING, Callable, Final, NamedTuple, Optional, Union

from dagster import _check as check
from dagster._annotations import beta, deprecated
from dagster._core.loader import LoadingContext
from dagster._utils.cached_method import cached_method

if TYPE_CHECKING:
    from dagster._core.definitions.base_asset_graph import BaseAssetGraph
    from dagster._core.definitions.events import (
        AssetKey,
        AssetKeyPartitionKey,
        AssetMaterialization,
        AssetObservation,
    )
    from dagster._core.event_api import EventLogRecord
    from dagster._core.events.log import EventLogEntry
    from dagster._core.instance import DagsterInstance
    from dagster._utils.caching_instance_queryer import CachingInstanceQueryer


class UnknownValue:
    pass


UNKNOWN_VALUE: Final[UnknownValue] = UnknownValue()


class DataVersion(
    NamedTuple(
        "_DataVersion",
        [("value", str)],
    )
):
    """Represents a data version for an asset.

    Args:
        value (str): An arbitrary string representing a data version.
    """

    def __new__(
        cls,
        value: str,
    ):
        return super().__new__(
            cls,
            value=check.str_param(value, "value"),
        )


@beta
class DataVersionsByPartition(
    NamedTuple(
        "_DataVersionsByPartition", [("data_versions_by_partition", Mapping[str, DataVersion])]
    )
):
    def __new__(
        cls,
        data_versions_by_partition: Mapping[str, Union[str, DataVersion]],
    ):
        check.dict_param(
            data_versions_by_partition,
            "data_versions_by_partition",
            key_type=str,
            value_type=(str, DataVersion),
        )
        return super().__new__(
            cls,
            data_versions_by_partition={
                partition: DataVersion(version) if isinstance(version, str) else version
                for partition, version in data_versions_by_partition.items()
            },
        )


DEFAULT_DATA_VERSION: Final[DataVersion] = DataVersion("INITIAL")
NULL_DATA_VERSION: Final[DataVersion] = DataVersion("NULL")
UNKNOWN_DATA_VERSION: Final[DataVersion] = DataVersion("UNKNOWN")
NULL_EVENT_POINTER: Final[str] = "NULL"


class DataProvenance(
    NamedTuple(
        "_DataProvenance",
        [
            ("code_version", str),
            ("input_data_versions", Mapping["AssetKey", DataVersion]),
            ("input_storage_ids", Mapping["AssetKey", Optional[int]]),
            ("is_user_provided", bool),
        ],
    )
):
    """Provenance information for an asset materialization.

    Args:
        code_version (str): The code version of the op that generated a materialization.
        input_data_versions (Mapping[AssetKey, DataVersion]): The data versions of the
            inputs used to generate a materialization.
        is_user_provided (bool): True if the data version was provided directly by the user, false
            if it was generated by Dagster.
    """

    def __new__(
        cls,
        code_version: str,
        input_data_versions: Mapping["AssetKey", DataVersion],
        input_storage_ids: Mapping["AssetKey", Optional[int]],
        is_user_provided: bool,
    ):
        from dagster._core.definitions.events import AssetKey

        return super().__new__(
            cls,
            code_version=check.str_param(code_version, "code_version"),
            input_data_versions=check.mapping_param(
                input_data_versions,
                "input_data_versions",
                key_type=AssetKey,
                value_type=DataVersion,
            ),
            input_storage_ids=check.mapping_param(
                input_storage_ids,
                "input_storage_ids",
                key_type=AssetKey,
            ),
            is_user_provided=check.bool_param(is_user_provided, "is_user_provided"),
        )

    # In theory an asset key should always be absent/present in both input_data_versions and
    # input_storage_ids, but in the wild we've observed at least one instance
    # where this was not true, resulting in an error. This method can be used
    # as a defensive check.
    def has_input_asset(self, key: "AssetKey") -> bool:
        return key in self.input_data_versions and key in self.input_storage_ids

    @staticmethod
    def from_tags(tags: Mapping[str, str]) -> Optional["DataProvenance"]:
        code_version = tags.get(CODE_VERSION_TAG)
        if code_version is None:
            return None
        input_data_versions = {
            _asset_key_from_tag(k): DataVersion(v)
            for k, v in tags.items()
            if k.startswith(INPUT_DATA_VERSION_TAG_PREFIX)
            or k.startswith(_OLD_INPUT_DATA_VERSION_TAG_PREFIX)
        }
        input_storage_ids = {
            _asset_key_from_tag(k): int(v) if v != NULL_EVENT_POINTER else None
            for k, v in tags.items()
            if k.startswith(INPUT_EVENT_POINTER_TAG_PREFIX)
        }
        is_user_provided = tags.get(DATA_VERSION_IS_USER_PROVIDED_TAG) == "true"
        return DataProvenance(
            code_version, input_data_versions, input_storage_ids, is_user_provided
        )

    @property
    @deprecated(breaking_version="2.0", additional_warn_text="Use `input_data_versions` instead.")
    def input_logical_versions(self) -> Mapping["AssetKey", DataVersion]:
        return self.input_data_versions


def _asset_key_from_tag(tag: str) -> "AssetKey":
    from dagster._core.definitions.events import AssetKey

    # Everything after the 2nd slash is the asset key
    return AssetKey.from_user_string(tag.split("/", maxsplit=2)[-1])


# ########################
# ##### TAG KEYS
# ########################

DATA_VERSION_TAG: Final[str] = "dagster/data_version"
_OLD_DATA_VERSION_TAG: Final[str] = "dagster/logical_version"
CODE_VERSION_TAG: Final[str] = "dagster/code_version"
INPUT_DATA_VERSION_TAG_PREFIX: Final[str] = "dagster/input_data_version"
_OLD_INPUT_DATA_VERSION_TAG_PREFIX: Final[str] = "dagster/input_logical_version"
INPUT_EVENT_POINTER_TAG_PREFIX: Final[str] = "dagster/input_event_pointer"
DATA_VERSION_IS_USER_PROVIDED_TAG = "dagster/data_version_is_user_provided"


def read_input_data_version_from_tags(
    tags: Mapping[str, str], input_key: "AssetKey"
) -> Optional[DataVersion]:
    value = tags.get(
        get_input_data_version_tag(input_key, prefix=INPUT_DATA_VERSION_TAG_PREFIX)
    ) or tags.get(get_input_data_version_tag(input_key, prefix=_OLD_INPUT_DATA_VERSION_TAG_PREFIX))
    return DataVersion(value) if value is not None else None


def get_input_data_version_tag(
    input_key: "AssetKey", prefix: str = INPUT_DATA_VERSION_TAG_PREFIX
) -> str:
    return f"{prefix}/{input_key.to_user_string()}"


def get_input_event_pointer_tag(input_key: "AssetKey") -> str:
    return f"{INPUT_EVENT_POINTER_TAG_PREFIX}/{input_key.to_user_string()}"


# ########################
# ##### COMPUTE / EXTRACT
# ########################


def compute_logical_data_version(
    code_version: Union[str, UnknownValue],
    input_data_versions: Mapping["AssetKey", DataVersion],
) -> DataVersion:
    """Compute a data version for a value as a hash of input data versions and code version.

    Args:
        code_version (str): The code version of the computation.
        input_data_versions (Mapping[AssetKey, DataVersion]): The data versions of the inputs.

    Returns:
        DataVersion: The computed version as a `DataVersion`.
    """
    from dagster._core.definitions.events import AssetKey

    check.inst_param(code_version, "code_version", (str, UnknownValue))
    check.mapping_param(
        input_data_versions, "input_versions", key_type=AssetKey, value_type=DataVersion
    )

    if (
        isinstance(code_version, UnknownValue)
        or UNKNOWN_DATA_VERSION in input_data_versions.values()
    ):
        return UNKNOWN_DATA_VERSION

    ordered_input_versions = [
        input_data_versions[k] for k in sorted(input_data_versions.keys(), key=str)
    ]
    all_inputs = (code_version, *(v.value for v in ordered_input_versions))

    hash_sig = sha256()
    hash_sig.update(bytearray("".join(all_inputs), "utf8"))
    return DataVersion(hash_sig.hexdigest())


def extract_data_version_from_entry(
    entry: "EventLogEntry",
) -> Optional[DataVersion]:
    tags = entry.tags or {}
    value = tags.get(DATA_VERSION_TAG) or tags.get(_OLD_DATA_VERSION_TAG)
    return None if value is None else DataVersion(value)


def extract_data_provenance_from_entry(
    entry: "EventLogEntry",
) -> Optional[DataProvenance]:
    tags = entry.tags or {}
    return DataProvenance.from_tags(tags)


# ########################
# ##### STALENESS OPERATIONS
# ########################


class StaleStatus(Enum):
    MISSING = "MISSING"
    STALE = "STALE"
    FRESH = "FRESH"


@functools.total_ordering
class StaleCauseCategory(Enum):
    CODE = "CODE"
    DATA = "DATA"
    DEPENDENCIES = "DEPENDENCIES"

    def __lt__(self, other: "StaleCauseCategory"):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class StaleCause(
    NamedTuple(
        "_StaleCause",
        [
            ("key", "AssetKeyPartitionKey"),
            ("category", StaleCauseCategory),
            ("reason", str),
            ("dependency", Optional["AssetKeyPartitionKey"]),
            ("children", Optional[Sequence["StaleCause"]]),
        ],
    )
):
    def __new__(
        cls,
        key: Union["AssetKey", "AssetKeyPartitionKey"],
        category: StaleCauseCategory,
        reason: str,
        dependency: Optional[Union["AssetKey", "AssetKeyPartitionKey"]] = None,
        children: Optional[Sequence["StaleCause"]] = None,
    ):
        from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey

        return super().__new__(
            cls,
            AssetKeyPartitionKey(key) if isinstance(key, AssetKey) else key,
            category,
            reason,
            AssetKeyPartitionKey(dependency) if isinstance(dependency, AssetKey) else dependency,
            children,
        )

    @property
    def asset_key(self) -> "AssetKey":
        return self.key.asset_key

    @property
    def partition_key(self) -> Optional[str]:
        return self.key.partition_key

    @property
    def dependency_asset_key(self) -> Optional["AssetKey"]:
        return self.dependency.asset_key if self.dependency else None

    @property
    def dependency_partition_key(self) -> Optional[str]:
        return self.dependency.partition_key if self.dependency else None

    @property
    def sort_key(self) -> str:
        if not hasattr(self, "_sort_key"):
            self._sort_key = f"{self.key}/{self.dependency}" if self.dependency else str(self.key)
        return self._sort_key

    @property
    def dedupe_key(self) -> int:
        # subset of properties that are safe to hash
        safe_tup = (self.key, self.category, self.reason, self.dependency)
        return hash(safe_tup)


# If a partition has greater than this number of dependencies, we don't check
# this edge for updated data or propagate other stale causes through this edge.
# This constraint can be removed when we have thoroughly tested performance for
# large upstream partition counts.
SKIP_PARTITION_DATA_VERSION_DEPENDENCY_THRESHOLD = 100

# If an asset is self-dependent and has  greater than this number of partitions, we don't check the
# self-edge for updated data or propagate other stale causes through the edge. That is because the
# current logic will recurse to the first partition, potentially throwing a recursion error. This
# constraint can be removed when we have thoroughly tested performance for large partition counts on
# self-dependent assets.
SKIP_PARTITION_DATA_VERSION_SELF_DEPENDENCY_THRESHOLD = 100


class CachingStaleStatusResolver:
    """Used to resolve data version information. Avoids redundant database
    calls that would otherwise occur. Intended for use within the scope of a
    single "request" (e.g. GQL request, RunRequest resolution).
    """

    _instance: "DagsterInstance"
    _instance_queryer: Optional["CachingInstanceQueryer"]
    _asset_graph: Optional["BaseAssetGraph"]
    _asset_graph_load_fn: Optional[Callable[[], "BaseAssetGraph"]]

    def __init__(
        self,
        instance: "DagsterInstance",
        asset_graph: Union["BaseAssetGraph", Callable[[], "BaseAssetGraph"]],
        loading_context: LoadingContext,
        instance_queryer: Optional["CachingInstanceQueryer"] = None,
    ):
        from dagster._core.definitions.base_asset_graph import BaseAssetGraph

        self._instance = instance
        self._instance_queryer = instance_queryer
        self._loading_context = loading_context
        if isinstance(asset_graph, BaseAssetGraph):
            self._asset_graph = asset_graph
            self._asset_graph_load_fn = None
        else:
            self._asset_graph = None
            self._asset_graph_load_fn = asset_graph

    def get_status(self, key: "AssetKey", partition_key: Optional[str] = None) -> StaleStatus:
        from dagster._core.definitions.events import AssetKeyPartitionKey

        return self._get_status(key=AssetKeyPartitionKey(key, partition_key))

    def get_stale_causes(
        self, key: "AssetKey", partition_key: Optional[str] = None
    ) -> Sequence[StaleCause]:
        from dagster._core.definitions.events import AssetKeyPartitionKey

        return self._get_stale_causes(key=AssetKeyPartitionKey(key, partition_key))

    def get_stale_root_causes(
        self, key: "AssetKey", partition_key: Optional[str] = None
    ) -> Sequence[StaleCause]:
        from dagster._core.definitions.events import AssetKeyPartitionKey

        return self._get_stale_root_causes(key=AssetKeyPartitionKey(key, partition_key))

    def get_current_data_version(
        self, key: "AssetKey", partition_key: Optional[str] = None
    ) -> DataVersion:
        from dagster._core.definitions.events import AssetKeyPartitionKey

        return self._get_current_data_version(key=AssetKeyPartitionKey(key, partition_key))

    @cached_method
    def _get_status(self, key: "AssetKeyPartitionKey") -> StaleStatus:
        # The status loader does not support querying for the stale status of a
        # partitioned asset without specifying a partition, so we return here.
        asset = self.asset_graph.get(key.asset_key)
        if asset.is_partitioned and not key.partition_key:
            return StaleStatus.FRESH
        else:
            current_version = self._get_current_data_version(key=key)
            if current_version == NULL_DATA_VERSION:
                return StaleStatus.MISSING
            elif asset.is_external:
                return StaleStatus.FRESH
            else:
                causes = self._get_stale_causes(key=key)
                return StaleStatus.FRESH if len(causes) == 0 else StaleStatus.STALE

    @cached_method
    def _get_stale_causes(self, key: "AssetKeyPartitionKey") -> Sequence[StaleCause]:
        # Querying for the stale status of a partitioned asset without specifying a partition key
        # is strictly speaking undefined, but we return an empty list here (from which FRESH status
        # is inferred) for backcompat.
        asset = self.asset_graph.get(key.asset_key)
        if asset.is_partitioned and not key.partition_key:
            return []
        elif asset.is_external:
            return []
        else:
            current_version = self._get_current_data_version(key=key)
            if current_version == NULL_DATA_VERSION:
                return []
            else:
                return sorted(
                    self._get_stale_causes_materialized(key=key), key=lambda cause: cause.sort_key
                )

    def _is_dep_updated(self, provenance: DataProvenance, dep_key: "AssetKeyPartitionKey") -> bool:
        dep_asset = self.asset_graph.get(dep_key.asset_key)
        if dep_key.partition_key is None:
            current_data_version = self._get_current_data_version(key=dep_key)
            return provenance.input_data_versions[dep_key.asset_key] != current_data_version
        else:
            cursor = provenance.input_storage_ids[dep_key.asset_key]
            updated_record = self._instance.get_latest_data_version_record(
                dep_key.asset_key,
                dep_asset.is_external,
                dep_key.partition_key,
                after_cursor=cursor,
            )
            if updated_record:
                previous_record = self._instance.get_latest_data_version_record(
                    dep_key.asset_key,
                    dep_asset.is_external,
                    dep_key.partition_key,
                    before_cursor=cursor + 1 if cursor else None,
                )
                previous_version = (
                    extract_data_version_from_entry(previous_record.event_log_entry)
                    if previous_record
                    else None
                )
                updated_version = extract_data_version_from_entry(updated_record.event_log_entry)
                return previous_version != updated_version
            else:
                return False

    def _get_stale_causes_materialized(self, key: "AssetKeyPartitionKey") -> Iterator[StaleCause]:
        from dagster._core.definitions.events import AssetKeyPartitionKey

        code_version = self.asset_graph.get(key.asset_key).code_version
        provenance = self._get_current_data_provenance(key=key)

        asset_deps = self.asset_graph.get(key.asset_key).parent_keys

        # only used if no provenance available
        materialization = check.not_none(self._get_latest_data_version_record(key=key))
        materialization_time = materialization.timestamp

        if provenance:
            if code_version and code_version != provenance.code_version:
                yield StaleCause(key, StaleCauseCategory.CODE, "has a new code version")

            removed_deps = set(provenance.input_data_versions.keys()) - set(asset_deps)
            for dep_key in removed_deps:
                yield StaleCause(
                    key,
                    StaleCauseCategory.DEPENDENCIES,
                    f"removed dependency on {dep_key.to_user_string()}",
                    AssetKeyPartitionKey(dep_key, None),
                )

        # If a partition has greater than or equal to SKIP_PARTITION_DATA_VERSION_DEPENDENCY_THRESHOLD of
        # dependencies, it is not included in partition_deps. This is for performance reasons. This
        # constraint can be removed when we have thoroughly tested performance for large upstream
        # partition counts.
        partition_deps = self._get_partition_dependencies(key=key)
        for dep_key in sorted(partition_deps):
            dep_asset = self.asset_graph.get(dep_key.asset_key)
            if provenance:
                if not provenance.has_input_asset(dep_key.asset_key):
                    yield StaleCause(
                        key,
                        StaleCauseCategory.DEPENDENCIES,
                        f"has a new dependency on {dep_key.asset_key.to_user_string()}",
                        dep_key,
                    )
                elif self._is_dep_updated(provenance, dep_key):
                    report_data_version = (
                        dep_asset.code_version is not None
                        or self._is_current_data_version_user_provided(key=dep_key)
                    )
                    yield StaleCause(
                        key,
                        StaleCauseCategory.DATA,
                        (
                            "has a new dependency data version"
                            if report_data_version
                            else "has a new dependency materialization"
                        ),
                        dep_key,
                        [
                            StaleCause(
                                dep_key,
                                StaleCauseCategory.DATA,
                                (
                                    # Assets with no defined code version will get a new data version on
                                    # every materialization, so we just report new materialization for
                                    # this case since the user likely hasn't engaged with data versions.
                                    "has a new data version"
                                    if report_data_version
                                    else "has a new materialization"
                                ),
                            )
                        ],
                    )
            # If no provenance and dep is a materializable asset, then use materialization
            # timestamps instead of versions this should be removable eventually since
            # provenance is on all newer materializations. If dep is a source, then we'll never
            # provide a stale reason here.
            elif not dep_asset.is_external:
                dep_materialization = self._get_latest_data_version_record(key=dep_key)
                if dep_materialization is None:
                    # The input must be new if it has no materialization
                    yield StaleCause(key, StaleCauseCategory.DATA, "has a new input", dep_key)
                elif dep_materialization.timestamp > materialization_time:
                    yield StaleCause(
                        key,
                        StaleCauseCategory.DATA,
                        "has a new dependency materialization",
                        dep_key,
                        [
                            StaleCause(
                                dep_key,
                                StaleCauseCategory.DATA,
                                "has a new materialization",
                            )
                        ],
                    )

    @cached_method
    def _get_stale_root_causes(self, key: "AssetKeyPartitionKey") -> Sequence[StaleCause]:
        candidates = self._get_stale_causes(key=key)
        visited = set()
        root_causes = []
        while candidates:
            next_candidates: list[StaleCause] = []
            for cause in candidates:
                if cause.dedupe_key not in visited:
                    if cause.children is None:
                        root_causes.append(cause)
                    else:
                        next_candidates.extend(cause.children)
                    visited.add(cause.dedupe_key)

            candidates = next_candidates
        return root_causes

    @property
    def asset_graph(self) -> "BaseAssetGraph":
        if self._asset_graph is None:
            self._asset_graph = check.not_none(self._asset_graph_load_fn)()
        return self._asset_graph

    # This is lazily constructed because it depends on the asset graph, which needs to be lazily
    # constructed for GQL performance reasons.
    @property
    def instance_queryer(self) -> "CachingInstanceQueryer":
        from dagster._utils.caching_instance_queryer import CachingInstanceQueryer

        if self._instance_queryer is None:
            self._instance_queryer = CachingInstanceQueryer(
                self._instance, self.asset_graph, self._loading_context
            )
        return self._instance_queryer

    @cached_method
    def _get_current_data_version(self, *, key: "AssetKeyPartitionKey") -> DataVersion:
        # Currently we can only use asset records, which are fetched in one shot, for non-source
        # assets. This is because the most recent AssetObservation is not stored on the AssetRecord.
        record = self._get_latest_data_version_record(key=key)
        if self.asset_graph.get(key.asset_key).is_external and record is None:
            return DEFAULT_DATA_VERSION
        elif record is None:
            return NULL_DATA_VERSION
        else:
            data_version = extract_data_version_from_entry(record.event_log_entry)
            return data_version or DEFAULT_DATA_VERSION

    @cached_method
    def _is_current_data_version_user_provided(self, *, key: "AssetKeyPartitionKey") -> bool:
        if self.asset_graph.get(key.asset_key).is_external:
            return True
        else:
            provenance = self._get_current_data_provenance(key=key)
            return provenance is not None and provenance.is_user_provided

    @cached_method
    def _get_current_data_provenance(
        self, *, key: "AssetKeyPartitionKey"
    ) -> Optional[DataProvenance]:
        record = self._get_latest_data_version_record(key=key)
        if record is None:
            return None
        else:
            return extract_data_provenance_from_entry(record.event_log_entry)

    # Volatility means that an asset is assumed to be constantly changing. We assume that observable
    # source assets are non-volatile, since the primary purpose of the observation function is to
    # determine if a source asset has changed. We assume that regular assets are volatile if they
    # are at the root of the graph (have no dependencies) or are downstream of a volatile asset.
    @cached_method
    def _is_volatile(self, *, key: "AssetKey") -> bool:
        asset = self.asset_graph.get(key)
        if asset.is_external:
            return asset.is_observable
        else:
            deps = asset.get(key).parent_keys
            return len(deps) == 0 or any(self._is_volatile(key=dep_key) for dep_key in deps)

    @cached_method
    def _get_latest_data_version_event(
        self, *, key: "AssetKeyPartitionKey"
    ) -> Optional[Union["AssetMaterialization", "AssetObservation"]]:
        record = self._get_latest_data_version_record(key=key)
        if record:
            entry = record.event_log_entry
            return entry.asset_materialization or entry.asset_observation
        else:
            return None

    @cached_method
    def _get_latest_data_version_record(
        self, key: "AssetKeyPartitionKey"
    ) -> Optional["EventLogRecord"]:
        return self.instance_queryer.get_latest_materialization_or_observation_record(
            asset_partition=key
        )

    # If a partition has greater than or equal to SKIP_PARTITION_DATA_VERSION_DEPENDENCY_THRESHOLD
    # of dependencies, or is downstream of a time window partition with an AllPartitionsMapping,
    # it is not included in partition_deps. This is for performance reasons. Besides this, the
    # logic here largely replicates `asset_graph.get_parents_partitions`.
    #
    # Similarly, If an asset is self-dependent and has greater than or equal to
    # SKIP_PARTITION_DATA_VERSION_SELF_DEPENDENCY_THRESHOLD partitions, we don't check the
    # self-edge for updated data or propagate other stale causes through the edge. That is because
    # the current logic will recurse to the first partition, potentially throwing a recursion error.
    @cached_method
    def _get_partition_dependencies(
        self, *, key: "AssetKeyPartitionKey"
    ) -> Sequence["AssetKeyPartitionKey"]:
        from dagster import AllPartitionMapping
        from dagster._core.definitions.events import AssetKeyPartitionKey
        from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsDefinition

        asset_deps = self.asset_graph.get(key.asset_key).parent_keys

        deps = []
        for dep_asset_key in asset_deps:
            dep_asset = self.asset_graph.get(dep_asset_key)
            if not dep_asset.is_partitioned:
                deps.append(AssetKeyPartitionKey(dep_asset_key, None))
            elif key.asset_key == dep_asset_key and self._exceeds_self_partition_limit(
                key.asset_key
            ):
                continue
            elif isinstance(
                dep_asset.partitions_def, TimeWindowPartitionsDefinition
            ) and isinstance(
                self.asset_graph.get_partition_mapping(key.asset_key, dep_asset_key),
                AllPartitionMapping,
            ):
                continue
            else:
                upstream_partition_keys = list(
                    self.asset_graph.get_parent_partition_keys_for_child(
                        key.partition_key,
                        dep_asset_key,
                        key.asset_key,
                        dynamic_partitions_store=self._instance,
                        current_time=self.instance_queryer.evaluation_time,
                    ).partitions_subset.get_partition_keys()
                )
                if len(upstream_partition_keys) < SKIP_PARTITION_DATA_VERSION_DEPENDENCY_THRESHOLD:
                    deps.extend(
                        [
                            AssetKeyPartitionKey(dep_asset_key, partition_key)
                            for partition_key in upstream_partition_keys
                        ]
                    )
        return deps

    def _exceeds_self_partition_limit(self, asset_key: "AssetKey") -> bool:
        return (
            check.not_none(self.asset_graph.get(asset_key).partitions_def).get_num_partitions(
                dynamic_partitions_store=self._instance
            )
            >= SKIP_PARTITION_DATA_VERSION_SELF_DEPENDENCY_THRESHOLD
        )



import contextlib
from collections.abc import Generator, Sequence
from typing import Optional, TypeVar

from pydantic import BaseModel, ValidationError, parse_obj_as

from dagster._core.errors import DagsterInvariantViolationError
from dagster._utils.source_position import (
    KeyPath,
    SourcePositionTree,
    ValueAndSourcePositionTree,
    populate_source_position_and_key_paths,
)
from dagster._utils.yaml_utils import parse_yaml_with_source_positions

T = TypeVar("T", bound=BaseModel)


def _parse_and_populate_model_with_annotated_errors(
    cls: type[T],
    obj_parse_root: ValueAndSourcePositionTree,
    obj_key_path_prefix: KeyPath,
) -> T:
    """Helper function to parse the Pydantic model from the parsed YAML object and populate source
    position information on the model and its sub-objects.

    Raises more helpful errors than Pydantic's default error messages, including the source position
    in the YAML file where the error occurred.

    Args:
        cls (Type[T]): The Pydantic model class to use for validation.
        obj_parse_root (ValueAndSourcePositionTree): The parsed YAML object to use for validation.
        file_root (Optional[ValueAndSourcePositionTree]): The root of the parsed YAML file, used for
            error reporting.
        obj_key_path_prefix (KeyPath): The path of keys that lead to the current object, used for
            both error reporting and populating source position information.
    """
    try:
        model = parse_obj_as(cls, obj_parse_root.value)
    except ValidationError as e:
        line_errors = []
        for error in e.errors():
            key_path_in_obj = list(error["loc"])
            source_position, source_position_path = (
                obj_parse_root.source_position_tree.lookup_closest_and_path(key_path_in_obj, None)
            )

            file_key_path: KeyPath = list(obj_key_path_prefix) + key_path_in_obj
            file_key_path_str = ".".join(str(part) for part in file_key_path)
            ctx = {
                **error.get("ctx", {}),
                "source_position": source_position,
                "source_position_path": source_position_path,
            }
            line_errors.append(
                {**error, "loc": [file_key_path_str + " at " + str(source_position)], "ctx": ctx}
            )

            file_key_path: KeyPath = list(obj_key_path_prefix) + key_path_in_obj
            file_key_path_str = ".".join(str(part) for part in file_key_path)

        raise ValidationError.from_exception_data(
            title=e.title,
            line_errors=line_errors,
            input_type="json",
            hide_input=False,
        ) from None

    populate_source_position_and_key_paths(
        model, obj_parse_root.source_position_tree, obj_key_path_prefix
    )
    return model


def parse_yaml_file_to_pydantic(cls: type[T], src: str, filename: str = "<string>") -> T:
    """Parse the YAML source and create a Pydantic model instance from it.

    Attaches source position information to the `_source_position_and_key_path` attribute of the
    Pydantic model instance and sub-objects.

    Args:
        cls (type[T]): The Pydantic model class to use for validation.
        src (str): The YAML source string to be parsed.
        filename (str): The filename associated with the YAML source, used for error reporting.
            Defaults to "<string>" if not provided.

    Returns:
        T: An instance of the Pydantic model class, with the `_source_position_and_key_path`
            attribute populated on it and all the objects inside it.

    Raises:
        ValidationError: If the YAML data does not conform to the Pydantic model schema. If using
            Pydantic2+, errors will include context information about the position in the document
            that the model corresponds to.
    """
    parsed = parse_yaml_with_source_positions(src, filename)
    return _parse_and_populate_model_with_annotated_errors(
        cls=cls, obj_parse_root=parsed, obj_key_path_prefix=[]
    )


@contextlib.contextmanager
def enrich_validation_errors_with_source_position(
    source_position_tree: SourcePositionTree, obj_key_path_prefix: KeyPath
) -> Generator[None, None, None]:
    err: Optional[ValidationError] = None
    try:
        yield
    except ValidationError as e:
        line_errors = []
        for error in e.errors():
            key_path_in_obj = list(error["loc"])
            source_position, source_position_path = source_position_tree.lookup_closest_and_path(
                key_path_in_obj, None
            )

            file_key_path: KeyPath = list(obj_key_path_prefix) + key_path_in_obj
            file_key_path_str = ".".join(str(part) for part in file_key_path)

            ctx = {
                **error.get("ctx", {}),
                "source_position": source_position,
                "source_position_path": source_position_path,
            }
            line_errors.append(
                {**error, "loc": [file_key_path_str + " at " + str(source_position)], "ctx": ctx}
            )

        err = ValidationError.from_exception_data(
            title=e.title,
            line_errors=line_errors,
            input_type="json",
            hide_input=False,
        )
    finally:
        if err:
            raise err from None


def parse_yaml_file_to_pydantic_sequence(
    cls: type[T], src: str, filename: str = "<string>"
) -> Sequence[T]:
    """Parse the YAML source and create a list of Pydantic model instances from it.

    Attaches source position information to the `_source_position_and_key_path` attribute of the
    Pydantic model instance and sub-objects.

    Args:
        cls (type[T]): The Pydantic model class to use for validation.
        src (str): The YAML source string to be parsed.
        filename (str): The filename associated with the YAML source, used for error reporting.
            Defaults to "<string>" if not provided.

    Returns:
        T: A list of instances of the Pydantic model class, with the `_source_position_and_key_path`
            attribute populated on each list element and all the objects inside them.

    Raises:
        ValidationError: If the YAML data does not conform to the Pydantic model schema. If using
            Pydantic2+, errors will include context information about the position in the document
            that the model corresponds to.
    """
    parsed = parse_yaml_with_source_positions(src, filename)

    if not isinstance(parsed.value, list):
        raise DagsterInvariantViolationError(
            f"Error parsing YAML file {filename}: Expected a list of objects at document root, but got {type(parsed.value)}"
        )

    results = []
    for i, entry in enumerate(parsed.value):
        results.append(
            _parse_and_populate_model_with_annotated_errors(
                cls=cls,
                obj_parse_root=ValueAndSourcePositionTree(
                    value=entry, source_position_tree=parsed.source_position_tree.children[i]
                ),
                obj_key_path_prefix=[i],
            )
        )
    return results



"""add concurrency limits table

Revision ID: 46b412388816
Revises: ec80dd91891a
Create Date: 2023-12-01 14:35:47.622154

"""

import sqlalchemy as db
from alembic import op
from dagster._core.storage.migration.utils import has_table
from dagster._core.storage.sql import MySQLCompatabilityTypes, get_sql_current_timestamp
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision = "46b412388816"
down_revision = "ec80dd91891a"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("concurrency_limits"):
        op.create_table(
            "concurrency_limits",
            db.Column(
                "id",
                db.BigInteger().with_variant(sqlite.INTEGER(), "sqlite"),
                primary_key=True,
                autoincrement=True,
            ),
            db.Column(
                "concurrency_key", MySQLCompatabilityTypes.UniqueText, nullable=False, unique=True
            ),
            db.Column("limit", db.Integer, nullable=False),
            db.Column("update_timestamp", db.DateTime, server_default=get_sql_current_timestamp()),
            db.Column("create_timestamp", db.DateTime, server_default=get_sql_current_timestamp()),
        )


def downgrade():
    if has_table("concurrency_limits"):
        op.drop_table("concurrency_limits")



import os
import signal
import subprocess
import sys
import threading
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from io import TextIOWrapper
from subprocess import Popen
from time import sleep
from typing import Any, Callable, NamedTuple, Optional

import dagster._check as check
from dagster._core.errors import DagsterError
from dagster._serdes.serdes import deserialize_value, serialize_value, whitelist_for_serdes
from dagster._utils import send_interrupt
from dagster._utils.error import (
    ExceptionInfo,
    SerializableErrorInfo,
    serializable_error_info_from_exc_info,
)


def write_unary_input(input_file: str, obj: NamedTuple) -> None:
    check.str_param(input_file, "input_file")
    check.not_none_param(obj, "obj")
    with open(os.path.abspath(input_file), "w", encoding="utf8") as fp:
        fp.write(serialize_value(obj))


def read_unary_input(input_file: str) -> tuple[object, ...]:
    check.str_param(input_file, "input_file")
    with open(os.path.abspath(input_file), encoding="utf8") as fp:
        return deserialize_value(fp.read(), NamedTuple)


def ipc_write_unary_response(output_file: str, obj: NamedTuple) -> None:
    check.not_none_param(obj, "obj")
    with ipc_write_stream(output_file) as stream:
        stream.send(obj)


def read_unary_response(
    output_file: str, timeout: int = 30, ipc_process: "Optional[Popen[Any]]" = None
) -> Optional[NamedTuple]:
    messages = list(ipc_read_event_stream(output_file, timeout=timeout, ipc_process=ipc_process))
    check.invariant(len(messages) == 1)
    return messages[0]


@whitelist_for_serdes
class IPCStartMessage(NamedTuple("_IPCStartMessage", [])):
    def __new__(cls):
        return super().__new__(cls)


@whitelist_for_serdes
class IPCErrorMessage(
    NamedTuple(
        "_IPCErrorMessage",
        [("serializable_error_info", SerializableErrorInfo), ("message", Optional[str])],
    )
):
    """This represents a user error encountered during the IPC call. This indicates a business
    logic error, rather than a protocol. Consider this a "task failed successfully"
    use case.
    """

    def __new__(cls, serializable_error_info: SerializableErrorInfo, message: Optional[str]):
        return super().__new__(
            cls,
            serializable_error_info=check.inst_param(
                serializable_error_info, "serializable_error_info", SerializableErrorInfo
            ),
            message=check.opt_str_param(message, "message"),
        )


@whitelist_for_serdes
class IPCEndMessage(NamedTuple("_IPCEndMessage", [])):
    def __new__(cls):
        return super().__new__(cls)


class DagsterIPCProtocolError(DagsterError):
    """This indicates that something went wrong with the protocol. E.g. the
    process being called did not emit an IPCStartMessage first.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class FileBasedWriteStream:
    def __init__(self, file_path: str):
        check.str_param("file_path", file_path)
        self._file_path = file_path

    def send(self, dagster_named_tuple: NamedTuple) -> None:
        _send(self._file_path, dagster_named_tuple)

    def send_error(self, exc_info: ExceptionInfo, message: Optional[str] = None) -> None:
        _send_error(self._file_path, exc_info, message=message)


def _send(file_path: str, obj: NamedTuple) -> None:
    with open(os.path.abspath(file_path), "a+", encoding="utf8") as fp:
        fp.write(serialize_value(obj) + "\n")


def _send_error(file_path: str, exc_info: ExceptionInfo, message: Optional[str]) -> None:
    return _send(
        file_path,
        IPCErrorMessage(
            serializable_error_info=serializable_error_info_from_exc_info(exc_info), message=message
        ),
    )


@contextmanager
def ipc_write_stream(file_path: str) -> Iterator[FileBasedWriteStream]:
    check.str_param("file_path", file_path)
    _send(file_path, IPCStartMessage())
    try:
        yield FileBasedWriteStream(file_path)
    except Exception:
        _send_error(file_path, sys.exc_info(), message=None)
    finally:
        _send(file_path, IPCEndMessage())


def _process_line(file_pointer: TextIOWrapper, sleep_interval: float = 0.1) -> Optional[NamedTuple]:
    while True:
        line = file_pointer.readline()
        if line:
            return deserialize_value(line.rstrip(), NamedTuple)
        sleep(sleep_interval)


def _poll_process(ipc_process: "Optional[Popen[Any]]") -> None:
    if not ipc_process:
        return
    if ipc_process.poll() is not None:
        raise DagsterIPCProtocolError(
            f"Process exited with return code {ipc_process.returncode} while waiting for events"
        )


def ipc_read_event_stream(
    file_path: str, timeout: int = 30, ipc_process: "Optional[Popen[Any]]" = None
) -> Iterator[Optional[NamedTuple]]:
    # Wait for file to be ready
    sleep_interval = 0.1
    elapsed_time = 0
    while elapsed_time < timeout and not os.path.exists(file_path):
        _poll_process(ipc_process)
        elapsed_time += sleep_interval
        sleep(sleep_interval)

    if not os.path.exists(file_path):
        raise DagsterIPCProtocolError(
            f"Timeout: read stream has not received any data in {timeout} seconds"
        )

    with open(os.path.abspath(file_path), encoding="utf8") as file_pointer:
        message = _process_line(file_pointer)
        while elapsed_time < timeout and message is None:
            _poll_process(ipc_process)
            elapsed_time += sleep_interval
            sleep(sleep_interval)
            message = _process_line(file_pointer)

        # Process start message
        if not isinstance(message, IPCStartMessage):
            raise DagsterIPCProtocolError(
                f"Attempted to read stream at file {file_path}, but first message was not an "
                "IPCStartMessage"
            )

        message = _process_line(file_pointer)
        while not isinstance(message, IPCEndMessage):
            if message is None:
                _poll_process(ipc_process)
            yield message
            message = _process_line(file_pointer)


# Windows subprocess termination utilities. See here for why we send CTRL_BREAK_EVENT on Windows:
# https://stefan.sofa-rockers.org/2013/08/15/handling-sub-process-hierarchies-python-linux-os-x/


def open_ipc_subprocess(parts: Sequence[str], **kwargs: Any) -> "Popen[Any]":
    """Sets the correct flags to support graceful termination."""
    check.list_param(parts, "parts", str)

    creationflags = 0
    if sys.platform == "win32":
        creationflags = subprocess.CREATE_NEW_PROCESS_GROUP

        # pass_fds is not supported on Windows. Instead we set close_fds to False, which will allow
        # any inheritable file descriptors marked as inheritable to be inherited by the child
        # process.
        if kwargs.get("pass_fds"):
            del kwargs["pass_fds"]
            kwargs["close_fds"] = False

    return subprocess.Popen(
        parts,
        creationflags=creationflags,
        **kwargs,
    )


def interrupt_ipc_subprocess(proc: "Popen[Any]", sig: Optional[int] = None) -> None:
    """Send CTRL_BREAK on Windows, SIGINT on other platforms."""
    proc.send_signal(sig or get_default_interrupt_signal())


def interrupt_then_kill_ipc_subprocess(proc: "Popen[Any]", wait_time: int = 10) -> None:
    interrupt_ipc_subprocess(proc)
    try:
        proc.wait(timeout=wait_time)
    except subprocess.TimeoutExpired:
        proc.kill()


def interrupt_ipc_subprocess_pid(pid: int, sig: Optional[int] = None) -> None:
    """Send CTRL_BREAK_EVENT on Windows, SIGINT on other platforms."""
    os.kill(pid, sig or get_default_interrupt_signal())


def get_default_interrupt_signal() -> int:
    return signal.CTRL_BREAK_EVENT if sys.platform == "win32" else signal.SIGINT


# ########################
# ##### SHUTDOWN PIPE
# ########################

_PIPE_SHUTDOWN_INDICATOR = "SHUTDOWN"


def get_ipc_shutdown_pipe() -> tuple[int, int]:
    r_fd, w_fd = os.pipe()

    # On windows, convert fd to a Windows handle so it can be reliably passed across processes.
    if sys.platform == "win32":
        import msvcrt

        os.set_inheritable(r_fd, True)
        r_fd = msvcrt.get_osfhandle(r_fd)
    return r_fd, w_fd


@contextmanager
def monitor_ipc_shutdown_pipe(pipe_fd: int, handler: Callable[[], None]) -> Iterator[None]:
    """Monitor the passed in pipe file descriptor for the shutdown indicator message.
    When received, trigger the handler.

    Args:
        pipe_fd: The file descriptor of the pipe to monitor. Must be readable.
            If on windows, this is assumed to be a Windows handle rather than a regular file
            descriptor.
        handler: The handler to call when the shutdown indicator is received.
    """
    # On windows, we expect to receive a raw Windows handle rather than a regular file descriptor.
    # Convert to a file descriptor before reading.
    if sys.platform == "win32":
        import msvcrt

        pipe_fd = msvcrt.open_osfhandle(pipe_fd, os.O_RDONLY)

    break_event = threading.Event()

    def _watch_pipe_for_shutdown():
        with open(pipe_fd) as pipe:
            while not break_event.is_set():
                line = pipe.readline()
                if not line:  # EOF or pipe closed
                    break_event.set()
                elif _PIPE_SHUTDOWN_INDICATOR in line.strip():
                    break_event.set()
                    handler()

    # Start a background thread that watches the pipe
    monitor_thread = threading.Thread(target=_watch_pipe_for_shutdown, daemon=True)
    monitor_thread.start()

    try:
        yield
    finally:
        # Signal the thread to exit and wait for it to stop
        break_event.set()
        monitor_thread.join()


@contextmanager
def interrupt_on_ipc_shutdown_message(pipe_fd: int) -> Iterator[None]:
    """Monitor the passed in pipe file descriptor for the shutdown indicator message. Interrupt the
    current process when the message is received.

    Args:
        pipe_fd: The file descriptor of the pipe to monitor. Must be readable.
            If on windows, this is assumed to be raw Windows handle rather than a regular file
            descriptor.
    """
    # Important to use `send_interrupt` here rather than unconditionally sending a signal. Sending a
    # signal, even to the process itself, often has strange behavior on windows.
    with monitor_ipc_shutdown_pipe(pipe_fd, handler=lambda: send_interrupt()):
        yield


def send_ipc_shutdown_message(w_fd: int) -> None:
    os.write(w_fd, f"{_PIPE_SHUTDOWN_INDICATOR}\n".encode())
    os.close(w_fd)



import datetime
import os
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, NamedTuple, Optional  # noqa: UP035

from dagster._annotations import deprecated
from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.auto_materialize_rule import AutoMaterializeRule
from dagster._core.definitions.auto_materialize_rule_evaluation import (
    AutoMaterializeDecisionType,
    ParentUpdatedRuleEvaluationData,
    WaitingOnAssetsRuleEvaluationData,
)
from dagster._core.definitions.base_asset_graph import sort_key_for_asset_partition
from dagster._core.definitions.declarative_automation.legacy.valid_asset_subset import (
    ValidAssetSubset,
)
from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey
from dagster._core.definitions.freshness_based_auto_materialize import (
    freshness_evaluation_results_for_asset_key,
)
from dagster._core.definitions.multi_dimensional_partitions import MultiPartitionsDefinition
from dagster._core.definitions.time_window_partitions import (
    TimeWindow,
    TimeWindowPartitionsDefinition,
    get_time_partitions_def,
)
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.event_api import AssetRecordsFilter
from dagster._core.storage.dagster_run import IN_PROGRESS_RUN_STATUSES, RunsFilter
from dagster._core.storage.tags import AUTO_MATERIALIZE_TAG
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.schedules import cron_string_iterator, reverse_cron_string_iterator

if TYPE_CHECKING:
    from dagster._core.definitions.declarative_automation.automation_condition import (
        AutomationResult,
    )
    from dagster._core.definitions.declarative_automation.automation_context import (
        AutomationContext,
    )


@whitelist_for_serdes
class MaterializeOnRequiredForFreshnessRule(
    AutoMaterializeRule, NamedTuple("_MaterializeOnRequiredForFreshnessRule", [])
):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.MATERIALIZE

    @property
    def description(self) -> str:
        return "required to meet this or downstream asset's freshness policy"

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        true_subset, subsets_with_metadata = freshness_evaluation_results_for_asset_key(
            context.legacy_context.root_context
        )
        true_subset = context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
            true_subset
        )
        return AutomationResult(context, true_subset, subsets_with_metadata=subsets_with_metadata)


@whitelist_for_serdes
class MaterializeOnCronRule(
    AutoMaterializeRule,
    NamedTuple(
        "_MaterializeOnCronRule",
        [("cron_schedule", str), ("timezone", str), ("all_partitions", bool)],
    ),
):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.MATERIALIZE

    @property
    def description(self) -> str:
        return f"not materialized since last cron schedule tick of '{self.cron_schedule}' (timezone: {self.timezone})"

    def missed_cron_ticks(self, context: "AutomationContext") -> Sequence[datetime.datetime]:
        """Returns the cron ticks which have been missed since the previous cursor was generated."""
        # if it's the first time evaluating this rule, then just count the latest tick as missed
        if (
            not context.legacy_context.node_cursor
            or not context.legacy_context.previous_evaluation_timestamp
        ):
            previous_dt = next(
                reverse_cron_string_iterator(
                    end_timestamp=context.legacy_context.evaluation_time.timestamp(),
                    cron_string=self.cron_schedule,
                    execution_timezone=self.timezone,
                )
            )
            return [previous_dt]
        missed_ticks = []
        for dt in cron_string_iterator(
            start_timestamp=context.legacy_context.previous_evaluation_timestamp,
            cron_string=self.cron_schedule,
            execution_timezone=self.timezone,
        ):
            if dt > context.legacy_context.evaluation_time:
                break
            missed_ticks.append(dt)
        return missed_ticks

    def get_new_candidate_asset_partitions(
        self,
        context: "AutomationContext",
        missed_ticks: Sequence[datetime.datetime],
    ) -> AbstractSet[AssetKeyPartitionKey]:
        if not missed_ticks:
            return set()

        partitions_def = context.legacy_context.partitions_def
        if partitions_def is None:
            return {AssetKeyPartitionKey(context.legacy_context.asset_key)}

        # if all_partitions is set, then just return all partitions if any ticks have been missed
        if self.all_partitions:
            return {
                AssetKeyPartitionKey(context.legacy_context.asset_key, partition_key)
                for partition_key in partitions_def.get_partition_keys(
                    current_time=context.legacy_context.evaluation_time,
                    dynamic_partitions_store=context.legacy_context.instance_queryer,
                )
            }

        # for partitions_defs without a time component, just return the last partition if any ticks
        # have been missed
        time_partitions_def = get_time_partitions_def(partitions_def)
        if time_partitions_def is None:
            return {
                AssetKeyPartitionKey(
                    context.legacy_context.asset_key,
                    partitions_def.get_last_partition_key(
                        dynamic_partitions_store=context.legacy_context.instance_queryer
                    ),
                )
            }

        missed_time_partition_keys = filter(
            None,
            [
                time_partitions_def.get_last_partition_key(
                    current_time=missed_tick,
                    dynamic_partitions_store=context.legacy_context.instance_queryer,
                )
                for missed_tick in missed_ticks
            ],
        )
        # for multi partitions definitions, request to materialize all partitions for each missed
        # cron schedule tick
        if isinstance(partitions_def, MultiPartitionsDefinition):
            return {
                AssetKeyPartitionKey(context.legacy_context.asset_key, partition_key)
                for time_partition_key in missed_time_partition_keys
                for partition_key in partitions_def.get_multipartition_keys_with_dimension_value(
                    partitions_def.time_window_dimension.name,
                    time_partition_key,
                    dynamic_partitions_store=context.legacy_context.instance_queryer,
                )
            }
        else:
            return {
                AssetKeyPartitionKey(context.legacy_context.asset_key, time_partition_key)
                for time_partition_key in missed_time_partition_keys
            }

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        missed_ticks = self.missed_cron_ticks(context)
        new_asset_partitions = self.get_new_candidate_asset_partitions(context, missed_ticks)

        # if it's the first time evaluating this rule, must query for the actual subset that has
        # been materialized since the previous cron tick, as materializations may have happened
        # before the previous evaluation, which
        # `context.legacy_context.materialized_requested_or_discarded_since_previous_tick_subset` would not capture
        if context.legacy_context.node_cursor is None:
            new_asset_partitions -= ValidAssetSubset.coerce_from_subset(
                context.legacy_context.instance_queryer.get_asset_subset_updated_after_time(
                    asset_key=context.legacy_context.asset_key, after_time=missed_ticks[-1]
                ),
                context.partitions_def,
            ).asset_partitions

        asset_subset_to_request = ValidAssetSubset.from_asset_partitions_set(
            context.legacy_context.asset_key,
            context.legacy_context.partitions_def,
            new_asset_partitions,
        ) | (
            ValidAssetSubset.coerce_from_subset(
                context.legacy_context.previous_true_subset, context.legacy_context.partitions_def
            )
            - context.legacy_context.materialized_requested_or_discarded_since_previous_tick_subset
        )

        true_subset = context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
            asset_subset_to_request
        )
        return AutomationResult(context, true_subset=true_subset)


@whitelist_for_serdes
@deprecated(breaking_version="1.10.0")
class AutoMaterializeAssetPartitionsFilter(
    NamedTuple(
        "_AutoMaterializeAssetPartitionsFilter",
        [("latest_run_required_tags", Optional[Mapping[str, str]])],
    )
):
    """A filter that can be applied to an asset partition, during auto-materialize evaluation, and
    returns a boolean for whether it passes.

    Args:
        latest_run_required_tags (Optional[Sequence[str]]): `passes` returns
            True if the run responsible for the latest materialization of the asset partition
            has all of these tags.
    """

    @property
    def description(self) -> str:
        return f"latest run includes required tags: {self.latest_run_required_tags}"

    def passes(
        self,
        context: "AutomationContext",
        asset_partitions: Iterable[AssetKeyPartitionKey],
    ) -> Iterable[AssetKeyPartitionKey]:
        if self.latest_run_required_tags is None:
            return asset_partitions

        will_update_asset_partitions: set[AssetKeyPartitionKey] = set()
        storage_ids_to_fetch_by_key: dict[AssetKey, list[int]] = defaultdict(list)

        for asset_partition in asset_partitions:
            if context.legacy_context.will_update_asset_partition(asset_partition):
                will_update_asset_partitions.add(asset_partition)
            else:
                latest_storage_id = context.legacy_context.instance_queryer.get_latest_materialization_or_observation_storage_id(
                    asset_partition=asset_partition
                )
                if latest_storage_id is not None:
                    storage_ids_to_fetch_by_key[asset_partition.asset_key].append(latest_storage_id)

        asset_partitions_by_latest_run_id: dict[str, set[AssetKeyPartitionKey]] = defaultdict(set)

        step = int(os.getenv("DAGSTER_ASSET_DAEMON_RUN_TAGS_EVENT_FETCH_LIMIT", "1000"))

        for asset_key, storage_ids_to_fetch in storage_ids_to_fetch_by_key.items():
            for i in range(0, len(storage_ids_to_fetch), step):
                storage_ids = storage_ids_to_fetch[i : i + step]
                fetch_records = (
                    context.legacy_context.instance_queryer.instance.fetch_observations
                    if context.legacy_context.asset_graph.get(asset_key).is_observable
                    else context.legacy_context.instance_queryer.instance.fetch_materializations
                )
                for record in fetch_records(
                    records_filter=AssetRecordsFilter(
                        asset_key=asset_key,
                        storage_ids=storage_ids,
                    ),
                    limit=step,
                ).records:
                    asset_partitions_by_latest_run_id[record.run_id].add(
                        AssetKeyPartitionKey(asset_key, record.partition_key)
                    )

        run_ids_with_required_tags = set()

        if len(asset_partitions_by_latest_run_id) > 0:
            run_step = int(os.getenv("DAGSTER_ASSET_DAEMON_RUN_TAGS_RUN_FETCH_LIMIT", "1000"))

            required_tag_items = self.latest_run_required_tags.items()

            run_ids_to_fetch = list(asset_partitions_by_latest_run_id.keys())
            for i in range(0, len(run_ids_to_fetch), run_step):
                run_ids = run_ids_to_fetch[i : i + run_step]
                runs = context.legacy_context.instance_queryer.instance.get_runs(
                    filters=RunsFilter(run_ids=run_ids)
                )
                run_ids_with_required_tags.update(
                    {run.run_id for run in runs if required_tag_items <= run.tags.items()}
                )

        updated_partitions_with_required_tags = {
            asset_partition
            for run_id, run_id_asset_partitions in asset_partitions_by_latest_run_id.items()
            if run_id in run_ids_with_required_tags
            for asset_partition in run_id_asset_partitions
        }

        if (
            self.latest_run_required_tags.items()
            <= {
                AUTO_MATERIALIZE_TAG: "true",
                **context.legacy_context.auto_materialize_run_tags,
            }.items()
        ):
            return will_update_asset_partitions | updated_partitions_with_required_tags
        else:
            return updated_partitions_with_required_tags

    def __hash__(self):
        return hash(frozenset((self.latest_run_required_tags or {}).items()))


@whitelist_for_serdes
class MaterializeOnParentUpdatedRule(
    AutoMaterializeRule,
    NamedTuple(
        "_MaterializeOnParentUpdatedRule",
        [("updated_parent_filter", Optional[AutoMaterializeAssetPartitionsFilter])],
    ),
):
    def __new__(cls, updated_parent_filter: Optional[AutoMaterializeAssetPartitionsFilter] = None):
        return super().__new__(cls, updated_parent_filter=updated_parent_filter)

    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.MATERIALIZE

    @property
    def description(self) -> str:
        base = "upstream data has changed since latest materialization"
        if self.updated_parent_filter is not None:
            return f"{base} and matches filter '{self.updated_parent_filter.description}'"
        else:
            return base

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        """Evaluates the set of asset partitions of this asset whose parents have been updated,
        or will update on this tick.
        """
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        asset_partitions_by_updated_parents: Mapping[
            AssetKeyPartitionKey, set[AssetKeyPartitionKey]
        ] = defaultdict(set)
        asset_partitions_by_will_update_parents: Mapping[
            AssetKeyPartitionKey, set[AssetKeyPartitionKey]
        ] = defaultdict(set)

        subset_to_evaluate = context.legacy_context.parent_has_or_will_update_subset
        for asset_partition in subset_to_evaluate.asset_partitions:
            parent_asset_partitions = context.legacy_context.asset_graph.get_parents_partitions(
                dynamic_partitions_store=context.legacy_context.instance_queryer,
                current_time=context.legacy_context.instance_queryer.evaluation_time,
                asset_key=asset_partition.asset_key,
                partition_key=asset_partition.partition_key,
            ).parent_partitions

            updated_parent_asset_partitions = context.legacy_context.instance_queryer.get_parent_asset_partitions_updated_after_child(
                asset_partition=asset_partition,
                parent_asset_partitions=parent_asset_partitions,
                # do a precise check for updated parents, factoring in data versions, as long as
                # we're within reasonable limits on the number of partitions to check
                respect_materialization_data_versions=context.legacy_context.respect_materialization_data_versions
                and len(parent_asset_partitions) + subset_to_evaluate.size < 100,
                # ignore self-dependencies when checking for updated parents, to avoid historical
                # rematerializations from causing a chain of materializations to be kicked off
                ignored_parent_keys={context.legacy_context.asset_key},
            )
            for parent in updated_parent_asset_partitions:
                asset_partitions_by_updated_parents[parent].add(asset_partition)

            for parent in parent_asset_partitions:
                if context.legacy_context.will_update_asset_partition(parent):
                    asset_partitions_by_will_update_parents[parent].add(asset_partition)

        updated_and_will_update_parents = (
            asset_partitions_by_updated_parents.keys()
            | asset_partitions_by_will_update_parents.keys()
        )
        filtered_updated_and_will_update_parents = (
            self.updated_parent_filter.passes(context, updated_and_will_update_parents)
            if self.updated_parent_filter
            else updated_and_will_update_parents
        )

        updated_parent_assets_by_asset_partition: dict[AssetKeyPartitionKey, set[AssetKey]] = (
            defaultdict(set)
        )
        will_update_parent_assets_by_asset_partition: dict[AssetKeyPartitionKey, set[AssetKey]] = (
            defaultdict(set)
        )

        for updated_or_will_update_parent in filtered_updated_and_will_update_parents:
            for child in asset_partitions_by_updated_parents.get(updated_or_will_update_parent, []):
                updated_parent_assets_by_asset_partition[child].add(
                    updated_or_will_update_parent.asset_key
                )

            for child in asset_partitions_by_will_update_parents.get(
                updated_or_will_update_parent, []
            ):
                will_update_parent_assets_by_asset_partition[child].add(
                    updated_or_will_update_parent.asset_key
                )

        asset_partitions_by_evaluation_data = defaultdict(set)
        for asset_partition in (
            updated_parent_assets_by_asset_partition.keys()
            | will_update_parent_assets_by_asset_partition.keys()
        ):
            asset_partitions_by_evaluation_data[
                ParentUpdatedRuleEvaluationData(
                    updated_asset_keys=frozenset(
                        updated_parent_assets_by_asset_partition.get(asset_partition, [])
                    ),
                    will_update_asset_keys=frozenset(
                        will_update_parent_assets_by_asset_partition.get(asset_partition, [])
                    ),
                ).frozen_metadata
            ].add(asset_partition)

        true_subset, subsets_with_metadata = (
            context.legacy_context.add_evaluation_data_from_previous_tick(
                asset_partitions_by_evaluation_data,
                ignore_subset=context.legacy_context.materialized_requested_or_discarded_since_previous_tick_subset,
            )
        )
        true_subset = context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
            true_subset
        )
        return AutomationResult(context, true_subset, subsets_with_metadata=subsets_with_metadata)


@whitelist_for_serdes
class MaterializeOnMissingRule(AutoMaterializeRule, NamedTuple("_MaterializeOnMissingRule", [])):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.MATERIALIZE

    @property
    def description(self) -> str:
        return "materialization is missing"

    def get_handled_subset(self, context: "AutomationContext") -> SerializableEntitySubset:
        """Returns the AssetSubset which has been handled (materialized, requested, or discarded).
        Accounts for cases in which the partitions definition may have changed between ticks.
        """
        previous_handled_subset = (
            context.legacy_context.node_cursor.get_structured_cursor(SerializableEntitySubset)
            if context.legacy_context.node_cursor
            else None
        )
        if (
            previous_handled_subset is None
            or not previous_handled_subset.is_compatible_with_partitions_def(context.partitions_def)
        ):
            previous_handled_subset = (
                context.legacy_context.instance_queryer.get_materialized_asset_subset(
                    asset_key=context.legacy_context.asset_key
                )
            )

        return (
            context.legacy_context.materialized_since_previous_tick_subset
            | context.legacy_context.previous_tick_requested_subset
            | previous_handled_subset
        )

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        """Evaluates the set of asset partitions for this asset which are missing and were not
        previously discarded.
        """
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        if (
            context.legacy_context.asset_key
            in context.legacy_context.asset_graph.root_executable_asset_keys
        ):
            handled_subset = self.get_handled_subset(context)
            unhandled_candidates = (
                context.legacy_context.candidate_subset
                & ValidAssetSubset.coerce_from_subset(
                    handled_subset, context.legacy_context.partitions_def
                ).inverse(
                    context.legacy_context.partitions_def,
                    context.legacy_context.evaluation_time,
                    context.legacy_context.instance_queryer,
                )
                if handled_subset.size > 0
                else context.legacy_context.candidate_subset
            )
        else:
            # to avoid causing potential perf issues, we are maintaing the previous behavior of
            # only marking a non-root asset as missing if at least one of its parents has updated
            # since the previous tick.
            # on top of this, we also check the latest time partition of the asset to see if it's
            # missing on the tick that it pops into existence
            asset_partitions_to_evaluate = (
                context.legacy_context.candidate_parent_has_or_will_update_subset.asset_partitions
            )
            if isinstance(context.legacy_context.partitions_def, TimeWindowPartitionsDefinition):
                last_partition_key = context.legacy_context.partitions_def.get_last_partition_key(
                    current_time=context.legacy_context.evaluation_time
                )
                previous_last_partition_key = (
                    context.legacy_context.partitions_def.get_last_partition_key(
                        current_time=datetime.datetime.fromtimestamp(
                            context.legacy_context.previous_evaluation_timestamp or 0,
                            tz=datetime.timezone.utc,
                        )
                    )
                )
                if last_partition_key != previous_last_partition_key:
                    asset_partitions_to_evaluate |= {
                        AssetKeyPartitionKey(context.legacy_context.asset_key, last_partition_key)
                    }

            handled_subset = None
            unhandled_candidates = (
                ValidAssetSubset.from_asset_partitions_set(
                    context.legacy_context.asset_key,
                    context.legacy_context.partitions_def,
                    {
                        ap
                        for ap in asset_partitions_to_evaluate
                        if not context.legacy_context.instance_queryer.asset_partition_has_materialization_or_observation(
                            ap
                        )
                    },
                )
                | context.legacy_context.previous_true_subset
            ) - context.legacy_context.previous_tick_requested_subset

        return AutomationResult(
            context,
            true_subset=context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
                unhandled_candidates
            ),
            # we keep track of the handled subset instead of the unhandled subset because new
            # partitions may spontaneously jump into existence at any time
            structured_cursor=handled_subset,
        )


@whitelist_for_serdes
class SkipOnParentOutdatedRule(AutoMaterializeRule, NamedTuple("_SkipOnParentOutdatedRule", [])):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.SKIP

    @property
    def description(self) -> str:
        return "waiting on upstream data to be up to date"

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        asset_partitions_by_evaluation_data = defaultdict(set)

        # only need to evaluate net-new candidates and candidates whose parents have changed
        subset_to_evaluate = (
            context.legacy_context.candidates_not_evaluated_on_previous_tick_subset
            | context.legacy_context.candidate_parent_has_or_will_update_subset
        )
        for candidate in subset_to_evaluate.asset_partitions:
            outdated_ancestors = set()
            # find the root cause of why this asset partition's parents are outdated (if any)
            for (
                parent
            ) in context.legacy_context.get_parents_that_will_not_be_materialized_on_current_tick(
                asset_partition=candidate
            ):
                if context.legacy_context.instance_queryer.have_ignorable_partition_mapping_for_outdated(
                    candidate.asset_key, parent.asset_key
                ):
                    continue
                outdated_ancestors.update(
                    context.legacy_context.instance_queryer.get_outdated_ancestors(
                        asset_partition=parent
                    )
                )
            if outdated_ancestors:
                asset_partitions_by_evaluation_data[
                    WaitingOnAssetsRuleEvaluationData(frozenset(outdated_ancestors)).frozen_metadata
                ].add(candidate)

        true_subset, subsets_with_metadata = (
            context.legacy_context.add_evaluation_data_from_previous_tick(
                asset_partitions_by_evaluation_data, ignore_subset=subset_to_evaluate
            )
        )
        true_subset = context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
            true_subset
        )
        return AutomationResult(context, true_subset, subsets_with_metadata=subsets_with_metadata)


@whitelist_for_serdes
class SkipOnParentMissingRule(AutoMaterializeRule, NamedTuple("_SkipOnParentMissingRule", [])):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.SKIP

    @property
    def description(self) -> str:
        return "waiting on upstream data to be present"

    def evaluate_for_asset(
        self,
        context: "AutomationContext",
    ) -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        asset_partitions_by_evaluation_data = defaultdict(set)

        # only need to evaluate net-new candidates and candidates whose parents have changed
        subset_to_evaluate = (
            context.legacy_context.candidates_not_evaluated_on_previous_tick_subset
            | context.legacy_context.candidate_parent_has_or_will_update_subset
        )
        for candidate in subset_to_evaluate.asset_partitions:
            missing_parent_asset_keys = set()
            for (
                parent
            ) in context.legacy_context.get_parents_that_will_not_be_materialized_on_current_tick(
                asset_partition=candidate
            ):
                # ignore missing or unexecutable assets, which will never have a materialization or
                # observation
                if not (
                    context.legacy_context.asset_graph.has(parent.asset_key)
                    and context.legacy_context.asset_graph.get(parent.asset_key).is_executable
                ):
                    continue
                if not context.legacy_context.instance_queryer.asset_partition_has_materialization_or_observation(
                    parent
                ):
                    missing_parent_asset_keys.add(parent.asset_key)
            if missing_parent_asset_keys:
                asset_partitions_by_evaluation_data[
                    WaitingOnAssetsRuleEvaluationData(
                        frozenset(missing_parent_asset_keys)
                    ).frozen_metadata
                ].add(candidate)

        true_subset, subsets_with_metadata = (
            context.legacy_context.add_evaluation_data_from_previous_tick(
                asset_partitions_by_evaluation_data, ignore_subset=subset_to_evaluate
            )
        )
        true_subset = context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
            true_subset
        )
        return AutomationResult(context, true_subset, subsets_with_metadata=subsets_with_metadata)


@whitelist_for_serdes
class SkipOnNotAllParentsUpdatedRule(
    AutoMaterializeRule,
    NamedTuple(
        "_SkipOnNotAllParentsUpdatedRule", [("require_update_for_all_parent_partitions", bool)]
    ),
):
    """An auto-materialize rule that enforces that an asset can only be materialized if all parents
    have been materialized since the asset's last materialization.

    Args:
        require_update_for_all_parent_partitions (Optional[bool]): Applies only to an unpartitioned
            asset or an asset partition that depends on more than one partition in any upstream asset.
            If true, requires all upstream partitions in each upstream asset to be materialized since
            the downstream asset's last materialization in order to update it. If false, requires at
            least one upstream partition in each upstream asset to be materialized since the downstream
            asset's last materialization in order to update it. Defaults to false.
    """

    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.SKIP

    @property
    def description(self) -> str:
        if self.require_update_for_all_parent_partitions is False:
            return "waiting on upstream data to be updated"
        else:
            return "waiting until all upstream partitions are updated"

    def evaluate_for_asset(
        self,
        context: "AutomationContext",
    ) -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        asset_partitions_by_evaluation_data = defaultdict(set)

        # only need to evaluate net-new candidates and candidates whose parents have changed
        subset_to_evaluate = (
            context.legacy_context.candidates_not_evaluated_on_previous_tick_subset
            | context.legacy_context.candidate_parent_has_or_will_update_subset
        )
        for candidate in subset_to_evaluate.asset_partitions:
            parent_partitions = context.legacy_context.asset_graph.get_parents_partitions(
                context.legacy_context.instance_queryer,
                context.legacy_context.instance_queryer.evaluation_time,
                context.legacy_context.asset_key,
                candidate.partition_key,
            ).parent_partitions

            updated_parent_partitions = (
                context.legacy_context.instance_queryer.get_parent_asset_partitions_updated_after_child(
                    asset_partition=candidate,
                    parent_asset_partitions=parent_partitions,
                    respect_materialization_data_versions=context.legacy_context.respect_materialization_data_versions,
                    ignored_parent_keys=set(),
                )
                | context.legacy_context.parent_will_update_subset.asset_partitions
            )

            if self.require_update_for_all_parent_partitions:
                # All upstream partitions must be updated in order for the candidate to be updated
                non_updated_parent_keys = {
                    parent.asset_key for parent in parent_partitions - updated_parent_partitions
                }
            else:
                # At least one upstream partition in each upstream asset must be updated in order
                # for the candidate to be updated
                parent_asset_keys = context.legacy_context.asset_graph.get(
                    context.legacy_context.asset_key
                ).parent_keys
                updated_parent_keys = {ap.asset_key for ap in updated_parent_partitions}
                non_updated_parent_keys = parent_asset_keys - updated_parent_keys

            # do not require past partitions of this asset to be updated
            non_updated_parent_keys -= {context.legacy_context.asset_key}

            if non_updated_parent_keys:
                asset_partitions_by_evaluation_data[
                    WaitingOnAssetsRuleEvaluationData(
                        frozenset(non_updated_parent_keys)
                    ).frozen_metadata
                ].add(candidate)

        true_subset, subsets_with_metadata = (
            context.legacy_context.add_evaluation_data_from_previous_tick(
                asset_partitions_by_evaluation_data, ignore_subset=subset_to_evaluate
            )
        )
        true_subset = context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
            true_subset
        )
        return AutomationResult(context, true_subset, subsets_with_metadata=subsets_with_metadata)


@whitelist_for_serdes
class SkipOnNotAllParentsUpdatedSinceCronRule(
    AutoMaterializeRule,
    NamedTuple(
        "_SkipOnNotAllParentsUpdatedSinceCronRule",
        [("cron_schedule", str), ("timezone", str)],
    ),
):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.SKIP

    @property
    def description(self) -> str:
        return f"waiting until all upstream assets have updated since the last cron schedule tick of '{self.cron_schedule}' (timezone: {self.timezone})"

    def passed_time_window(self, context: "AutomationContext") -> TimeWindow:
        """Returns the window of time that has passed between the previous two cron ticks. All
        parent assets must contain all data from this time window in order for this asset to be
        materialized.
        """
        previous_ticks = reverse_cron_string_iterator(
            end_timestamp=context.legacy_context.evaluation_time.timestamp(),
            cron_string=self.cron_schedule,
            execution_timezone=self.timezone,
        )
        end_time = next(previous_ticks)
        start_time = next(previous_ticks)

        return TimeWindow(start=start_time, end=end_time)

    def get_parent_subset_updated_since_cron(
        self,
        context: "AutomationContext",
        parent_asset_key: AssetKey,
        passed_time_window: TimeWindow,
    ) -> ValidAssetSubset:
        """Returns the AssetSubset of a given parent asset that has been updated since the end of
        the previous cron tick. If a value for this parent asset was computed on the previous
        evaluation, and that evaluation happened within the same cron tick as the current evaluation,
        then this value will be calculated incrementally from the previous value to avoid expensive
        queries.
        """
        parent_partitions_def = context.asset_graph.get(parent_asset_key).partitions_def
        if (
            # first tick of evaluating this condition
            context.legacy_context.node_cursor is None
            or context.legacy_context.previous_evaluation_timestamp is None
            # This additional check is neccessary as it is possible for this cursor to be None
            # even if the previous state is not None in the case that this asset and none of its
            # parents have been detected as materialized at any point. While in theory this should
            # not cause issues, in past versions we were not properly capturing materializations
            # of external assets, causing this detection to be faulty. This ensures that we can
            # properly re-calculate this value even in the case that we've incorrectly considered
            # a parent as unmaterialized in the past.
            or context.legacy_context.previous_max_storage_id is None
            # new cron tick has happened since the previous tick
            or passed_time_window.end.timestamp()
            > context.legacy_context.previous_evaluation_timestamp
        ):
            return ValidAssetSubset.coerce_from_subset(
                context.legacy_context.instance_queryer.get_asset_subset_updated_after_time(
                    asset_key=parent_asset_key, after_time=passed_time_window.end
                ),
                parent_partitions_def,
            )
        else:
            # previous state still valid
            previous_parent_subsets = (
                context.legacy_context.node_cursor.get_structured_cursor(list) or []
            )
            previous_parent_subset = next(
                (s for s in previous_parent_subsets if s.key == parent_asset_key),
                ValidAssetSubset.empty(
                    parent_asset_key, context.asset_graph.get(parent_asset_key).partitions_def
                ),
            )

            # the set of asset partitions that have been updated since the previous evaluation
            new_parent_subset = ValidAssetSubset.coerce_from_subset(
                context.legacy_context.instance_queryer.get_asset_subset_updated_after_cursor(
                    asset_key=parent_asset_key,
                    after_cursor=context.legacy_context.previous_max_storage_id,
                    require_data_version_update=False,
                ),
                parent_partitions_def,
            )
            return new_parent_subset | previous_parent_subset

    def get_parent_subsets_updated_since_cron_by_key(
        self, context: "AutomationContext", passed_time_window: TimeWindow
    ) -> Mapping[AssetKey, ValidAssetSubset]:
        """Returns a mapping of parent asset keys to the AssetSubset of each parent that has been
        updated since the end of the previous cron tick. Does not compute this value for time-window
        partitioned parents, as their partitions encode the time windows they have processed.
        """
        updated_subsets_by_key = {}
        for parent_asset_key in context.legacy_context.asset_graph.get(
            context.legacy_context.asset_key
        ).parent_keys:
            # no need to incrementally calculate updated time-window partitions definitions, as
            # their partitions encode the time windows they have processed.
            if isinstance(
                context.legacy_context.asset_graph.get(parent_asset_key).partitions_def,
                TimeWindowPartitionsDefinition,
            ):
                continue
            updated_subsets_by_key[parent_asset_key] = self.get_parent_subset_updated_since_cron(
                context, parent_asset_key, passed_time_window
            )
        return updated_subsets_by_key

    def parent_updated_since_cron(
        self,
        context: "AutomationContext",
        passed_time_window: TimeWindow,
        parent_asset_key: AssetKey,
        child_asset_partition: AssetKeyPartitionKey,
        updated_parent_subset: ValidAssetSubset,
    ) -> bool:
        """Returns if, for a given child asset partition, the given parent asset been updated with
        information from the required time window.
        """
        parent_partitions_def = context.legacy_context.asset_graph.get(
            parent_asset_key
        ).partitions_def

        if isinstance(parent_partitions_def, TimeWindowPartitionsDefinition):
            # for time window partitions definitions, we simply assert that all time partitions that
            # were newly created between the previous cron ticks have been materialized
            required_parent_partitions = parent_partitions_def.get_partition_keys_in_time_window(
                time_window=passed_time_window
            )

            # for time window partitions definitions, we simply assert that all time partitions that
            return all(
                AssetKeyPartitionKey(parent_asset_key, partition_key)
                in context.legacy_context.instance_queryer.get_materialized_asset_subset(
                    asset_key=parent_asset_key
                )
                for partition_key in required_parent_partitions
            )
        # for all other partitions definitions, we assert that all parent partition keys have
        # been materialized since the previous cron tick
        else:
            if parent_partitions_def is None:
                non_updated_parent_asset_partitions = updated_parent_subset.inverse(
                    parent_partitions_def
                ).asset_partitions
            else:
                parent_subset = (
                    context.legacy_context.asset_graph.get_parent_partition_keys_for_child(
                        child_asset_partition.partition_key,
                        parent_asset_key,
                        child_asset_partition.asset_key,
                        context.legacy_context.instance_queryer,
                        context.legacy_context.evaluation_time,
                    ).partitions_subset
                )

                non_updated_parent_asset_partitions = (
                    ValidAssetSubset(key=parent_asset_key, value=parent_subset)
                    - updated_parent_subset
                ).asset_partitions

            return not any(
                not context.legacy_context.will_update_asset_partition(p)
                for p in non_updated_parent_asset_partitions
            )

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        passed_time_window = self.passed_time_window(context)
        has_new_passed_time_window = passed_time_window.end.timestamp() > (
            context.legacy_context.previous_evaluation_timestamp or 0
        )
        updated_subsets_by_key = self.get_parent_subsets_updated_since_cron_by_key(
            context, passed_time_window
        )

        # only need to evaluate net-new candidates and candidates whose parents have updated, unless
        # this is the first tick after a new cron schedule tick
        subset_to_evaluate = (
            (
                context.legacy_context.candidates_not_evaluated_on_previous_tick_subset
                | context.legacy_context.candidate_parent_has_or_will_update_subset
            )
            if not has_new_passed_time_window
            else context.legacy_context.candidate_subset
        )

        # the set of candidates for whom all parents have been updated since the previous cron tick
        all_parents_updated_subset = ValidAssetSubset.from_asset_partitions_set(
            context.legacy_context.asset_key,
            context.legacy_context.partitions_def,
            {
                candidate
                for candidate in subset_to_evaluate.asset_partitions
                if all(
                    self.parent_updated_since_cron(
                        context,
                        passed_time_window,
                        parent_asset_key,
                        candidate,
                        updated_subsets_by_key.get(
                            parent_asset_key,
                            ValidAssetSubset.empty(
                                parent_asset_key,
                                context.asset_graph.get(parent_asset_key).partitions_def,
                            ),
                        ),
                    )
                    for parent_asset_key in context.legacy_context.asset_graph.get(
                        candidate.asset_key
                    ).parent_keys
                )
            },
        )
        # if your parents were all updated since the previous cron tick on the previous evaluation,
        # that will still be true unless a new cron tick has happened since the previous evaluation
        if not has_new_passed_time_window:
            all_parents_updated_subset = (
                ValidAssetSubset.coerce_from_subset(
                    context.legacy_context.previous_candidate_subset,
                    context.legacy_context.partitions_def,
                )
                - context.legacy_context.previous_true_subset
            ) | all_parents_updated_subset

        return AutomationResult(
            context,
            true_subset=context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
                context.legacy_context.candidate_subset - all_parents_updated_subset
            ),
            structured_cursor=list(updated_subsets_by_key.values()),
        )


@whitelist_for_serdes
class SkipOnRequiredButNonexistentParentsRule(
    AutoMaterializeRule, NamedTuple("_SkipOnRequiredButNonexistentParentsRule", [])
):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.SKIP

    @property
    def description(self) -> str:
        return "required parent partitions do not exist"

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        asset_partitions_by_evaluation_data = defaultdict(set)

        subset_to_evaluate = (
            context.legacy_context.candidates_not_evaluated_on_previous_tick_subset
            | context.legacy_context.candidate_parent_has_or_will_update_subset
        )
        for candidate in subset_to_evaluate.asset_partitions:
            nonexistent_parent_partitions = (
                context.legacy_context.asset_graph.get_parents_partitions(
                    context.legacy_context.instance_queryer,
                    context.legacy_context.instance_queryer.evaluation_time,
                    candidate.asset_key,
                    candidate.partition_key,
                ).required_but_nonexistent_parents_partitions
            )

            nonexistent_parent_keys = {parent.asset_key for parent in nonexistent_parent_partitions}
            if nonexistent_parent_keys:
                asset_partitions_by_evaluation_data[
                    WaitingOnAssetsRuleEvaluationData(
                        frozenset(nonexistent_parent_keys)
                    ).frozen_metadata
                ].add(candidate)

        true_subset, subsets_with_metadata = (
            context.legacy_context.add_evaluation_data_from_previous_tick(
                asset_partitions_by_evaluation_data, ignore_subset=subset_to_evaluate
            )
        )
        true_subset = context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
            true_subset
        )
        return AutomationResult(context, true_subset, subsets_with_metadata=subsets_with_metadata)


@whitelist_for_serdes
class SkipOnBackfillInProgressRule(
    AutoMaterializeRule,
    NamedTuple("_SkipOnBackfillInProgressRule", [("all_partitions", bool)]),
):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.SKIP

    @property
    def description(self) -> str:
        if self.all_partitions:
            return "part of an asset targeted by an in-progress backfill"
        else:
            return "targeted by an in-progress backfill"

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        backfilling_subset = ValidAssetSubset.coerce_from_subset(
            # this backfilling subset is aware of the current partitions definitions, and so will
            # be valid
            (
                context.legacy_context.instance_queryer.get_active_backfill_target_asset_graph_subset()
            ).get_asset_subset(
                context.legacy_context.asset_key, context.legacy_context.asset_graph
            ),
            context.legacy_context.partitions_def,
        )

        if backfilling_subset.size == 0:
            true_subset = context.legacy_context.empty_subset()
        elif self.all_partitions:
            true_subset = context.legacy_context.candidate_subset
        else:
            true_subset = context.legacy_context.candidate_subset & backfilling_subset

        true_subset = context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
            true_subset
        )
        return AutomationResult(context, true_subset)


@whitelist_for_serdes
class DiscardOnMaxMaterializationsExceededRule(
    AutoMaterializeRule, NamedTuple("_DiscardOnMaxMaterializationsExceededRule", [("limit", int)])
):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.DISCARD

    @property
    def description(self) -> str:
        return f"exceeds {self.limit} materialization(s) per minute"

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        # the set of asset partitions which exceed the limit
        rate_limited_asset_partitions = set(
            sorted(
                context.legacy_context.candidate_subset.asset_partitions,
                key=lambda x: sort_key_for_asset_partition(context.legacy_context.asset_graph, x),
            )[self.limit :]
        )

        return AutomationResult(
            context,
            context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
                ValidAssetSubset.from_asset_partitions_set(
                    context.legacy_context.asset_key,
                    context.legacy_context.partitions_def,
                    rate_limited_asset_partitions,
                )
            ),
        )


@whitelist_for_serdes
class SkipOnRunInProgressRule(AutoMaterializeRule, NamedTuple("_SkipOnRunInProgressRule", [])):
    @property
    def decision_type(self) -> AutoMaterializeDecisionType:
        return AutoMaterializeDecisionType.SKIP

    @property
    def description(self) -> str:
        return "in-progress run for asset"

    def evaluate_for_asset(self, context: "AutomationContext") -> "AutomationResult":
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationResult,
        )

        if context.legacy_context.partitions_def is not None:
            raise DagsterInvariantViolationError(
                "SkipOnRunInProgressRule is currently only support for non-partitioned assets."
            )
        instance = context.legacy_context.instance_queryer.instance
        planned_materialization_info = (
            instance.event_log_storage.get_latest_planned_materialization_info(
                context.legacy_context.asset_key
            )
        )
        if planned_materialization_info:
            dagster_run = instance.get_run_by_id(planned_materialization_info.run_id)
            if dagster_run and dagster_run.status in IN_PROGRESS_RUN_STATUSES:
                return AutomationResult(
                    context,
                    context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
                        context.legacy_context.candidate_subset
                    ),
                )
        return AutomationResult(
            context,
            context.asset_graph_view.legacy_get_asset_subset_from_valid_subset(
                context.legacy_context.empty_subset()
            ),
        )



from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import TYPE_CHECKING

from dagster import _check as check
from dagster._annotations import public
from dagster._core.execution.plan.objects import StepFailureData, StepRetryData
from dagster._core.execution.retries import RetryMode
from dagster._utils.error import SerializableErrorInfo

if TYPE_CHECKING:
    from dagster._core.events import DagsterEvent
    from dagster._core.execution.context.system import IStepContext, PlanOrchestrationContext
    from dagster._core.execution.plan.plan import ExecutionPlan
    from dagster._core.execution.plan.state import KnownExecutionState


class Executor(ABC):
    @public
    @abstractmethod
    def execute(
        self, plan_context: "PlanOrchestrationContext", execution_plan: "ExecutionPlan"
    ) -> Iterator["DagsterEvent"]:
        """For the given context and execution plan, orchestrate a series of sub plan executions in a way that satisfies the whole plan being executed.

        Args:
            plan_context (PlanOrchestrationContext): The plan's orchestration context.
            execution_plan (ExecutionPlan): The plan to execute.

        Returns:
            A stream of dagster events.
        """

    @public
    @property
    @abstractmethod
    def retries(self) -> RetryMode:
        """Whether retries are enabled or disabled for this instance of the executor.

        Executors should allow this to be controlled via configuration if possible.

        Returns: RetryMode
        """

    def get_failure_or_retry_event_after_crash(
        self,
        step_context: "IStepContext",
        err_info: SerializableErrorInfo,
        known_state: "KnownExecutionState",
    ):
        from dagster._core.events import DagsterEvent

        # determine the retry policy for the step if needed
        retry_policy = step_context.op_retry_policy
        retry_state = known_state.get_retry_state()
        previous_attempt_count = retry_state.get_attempt_count(step_context.step.key)
        should_retry = (
            retry_policy
            and not step_context.retry_mode.disabled
            and previous_attempt_count < retry_policy.max_retries
        )

        if should_retry:
            return DagsterEvent.step_retry_event(
                step_context,
                StepRetryData(
                    error=err_info,
                    seconds_to_wait=check.not_none(retry_policy).calculate_delay(
                        previous_attempt_count + 1
                    ),
                ),
            )
        else:
            return DagsterEvent.step_failure_event(
                step_context=step_context,
                step_failure_data=StepFailureData(error=err_info, user_failure_data=None),
            )



import logging
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Optional, Tuple  # noqa: F401, UP035

import coloredlogs

from dagster import _seven
from dagster._config import Field
from dagster._core.definitions.logger_definition import LoggerDefinition, logger
from dagster._core.log_manager import (
    LOG_RECORD_EVENT_ATTR,
    LOG_RECORD_METADATA_ATTR,
    DagsterLogRecordMetadata,
)
from dagster._core.utils import coerce_valid_log_level
from dagster._serdes import pack_value
from dagster._utils.log import create_console_logger

if TYPE_CHECKING:
    from dagster._core.execution.context.logger import InitLoggerContext
    from dagster._core.instance import DagsterInstance


@logger(
    Field(
        {
            "log_level": Field(
                str,
                is_required=False,
                default_value="INFO",
                description="The logger's threshold.",
            ),
            "name": Field(
                str,
                is_required=False,
                default_value="dagster",
                description="The name of your logger.",
            ),
        },
        description="The default colored console logger.",
    ),
    description="The default colored console logger.",
)
def colored_console_logger(init_context: "InitLoggerContext") -> logging.Logger:
    """This logger provides support for sending Dagster logs to stdout in a colored format. It is
    included by default on jobs which do not otherwise specify loggers.
    """
    return create_console_logger(
        name=init_context.logger_config["name"],
        level=coerce_valid_log_level(init_context.logger_config["log_level"]),
    )


class JsonLogFormatter(logging.Formatter):
    def format(self, record):
        dict_to_dump = {}
        for k, v in record.__dict__.items():
            if k == LOG_RECORD_EVENT_ATTR:
                # Redundant with the "dagster_event" field under "dagster_meta"
                continue
            elif k == LOG_RECORD_METADATA_ATTR:
                # Events objects are not always JSON-serializable, so need to pack them first
                json_serializable_event = pack_value(v[LOG_RECORD_EVENT_ATTR])
                json_serializable_dagster_meta = DagsterLogRecordMetadata(
                    **{**v, "dagster_event": json_serializable_event}
                )
                dict_to_dump[LOG_RECORD_METADATA_ATTR] = json_serializable_dagster_meta
            else:
                dict_to_dump[k] = v

        return _seven.json.dumps(dict_to_dump)


@logger(
    Field(
        {
            "log_level": Field(
                str,
                is_required=False,
                default_value="INFO",
                description="The logger's threshold.",
            ),
            "name": Field(
                str,
                is_required=False,
                default_value="dagster",
                description="The name of your logger.",
            ),
        },
        description="A JSON-formatted console logger.",
    ),
    description="A JSON-formatted console logger.",
)
def json_console_logger(init_context: "InitLoggerContext") -> logging.Logger:
    """This logger provides support for sending Dagster logs to stdout in json format.

    Example:
        .. code-block:: python

            from dagster import op, job
            from dagster.loggers import json_console_logger

            @op
            def hello_op(context):
                context.log.info('Hello, world!')
                context.log.error('This is an error')

            @job(logger_defs={'json_logger': json_console_logger})])
            def json_logged_job():
                hello_op()

    """
    level = coerce_valid_log_level(init_context.logger_config["log_level"])
    name = init_context.logger_config["name"]

    klass = logging.getLoggerClass()
    logger_ = klass(name, level=level)

    handler = coloredlogs.StandardErrorHandler()

    handler.setFormatter(JsonLogFormatter())
    logger_.addHandler(handler)

    return logger_


def default_system_loggers(
    instance: Optional["DagsterInstance"],
) -> Sequence[tuple["LoggerDefinition", Mapping[str, object]]]:
    """If users don't provide configuration for any loggers, we instantiate these loggers with the
    default config.

    Returns:
        List[Tuple[LoggerDefinition, dict]]: Default loggers and their associated configs.
    """
    log_level = instance.python_log_level if (instance and instance.python_log_level) else "DEBUG"
    return [(colored_console_logger, {"name": "dagster", "log_level": log_level})]


def default_loggers() -> Mapping[str, "LoggerDefinition"]:
    return {"console": colored_console_logger}



# -*- coding: utf-8 -*-
# Vendored from https://github.com/dateutil/dateutil/releases/tag/2.9.0
# Any changes are marked with "# CHANGED IN VENDORED VERSION" (e.g.
# replacing absolute dateutil imports with relative imports)
import sys


# CHANGED IN VENDORED VERSION
__version__ = "2.9.0-vendored"

__all__ = ['easter', 'parser', 'relativedelta', 'rrule', 'tz',
           'utils', 'zoneinfo']

def __getattr__(name):
    import importlib

    if name in __all__:
        return importlib.import_module("." + name, __name__)
    raise AttributeError(
        "module {!r} has not attribute {!r}".format(__name__, name)
    )


def __dir__():
    # __dir__ should include all the lazy-importable modules as well.
    return [x for x in globals() if x not in sys.modules] + __all__



"""0.10.0 create new schedule tables.

Revision ID: 493871843165
Revises: 942138e33bf9
Create Date: 2021-01-13 14:43:03.678784

"""

from dagster._core.storage.migration.utils import create_0_10_0_schedule_tables

# revision identifiers, used by Alembic.
revision = "493871843165"
down_revision = "942138e33bf9"
branch_labels = None
depends_on = None


def upgrade():
    create_0_10_0_schedule_tables()


def downgrade():
    pass



import inspect
import os
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

import dagster._check as check
from dagster._annotations import beta, public
from dagster._core.definitions.metadata.metadata_set import (
    NamespacedMetadataSet as NamespacedMetadataSet,
    TableMetadataSet as TableMetadataSet,
)
from dagster._core.definitions.metadata.metadata_value import MetadataValue
from dagster._model import DagsterModel
from dagster._serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.definitions.assets import AssetsDefinition, AssetSpec, SourceAsset
    from dagster._core.definitions.cacheable_assets import CacheableAssetsDefinition

DEFAULT_SOURCE_FILE_KEY = "asset_definition"


@beta
@whitelist_for_serdes
class LocalFileCodeReference(DagsterModel):
    """Represents a local file source location."""

    file_path: str
    line_number: Optional[int] = None
    label: Optional[str] = None


@beta
@whitelist_for_serdes
class UrlCodeReference(DagsterModel):
    """Represents a source location which points at a URL, for example
    in source control.
    """

    url: str
    label: Optional[str] = None


@beta
@whitelist_for_serdes
class CodeReferencesMetadataValue(DagsterModel, MetadataValue["CodeReferencesMetadataValue"]):
    """Metadata value type which represents source locations (locally or otherwise)
    of the asset in question. For example, the file path and line number where the
    asset is defined.

    Args:
        sources (List[Union[LocalFileCodeReference, SourceControlCodeReference]]):
            A list of code references for the asset, such as file locations or
            references to source control.
    """

    code_references: list[Union[LocalFileCodeReference, UrlCodeReference]]

    @property
    def value(self) -> "CodeReferencesMetadataValue":
        return self


def local_source_path_from_fn(fn: Callable[..., Any]) -> Optional[LocalFileCodeReference]:
    cwd = os.getcwd()

    origin_file = os.path.abspath(os.path.join(cwd, inspect.getsourcefile(fn)))  # type: ignore
    origin_file = check.not_none(origin_file)
    origin_line = inspect.getsourcelines(fn)[1]

    return LocalFileCodeReference(file_path=origin_file, line_number=origin_line)


class CodeReferencesMetadataSet(NamespacedMetadataSet):
    """Metadata entries that apply to asset definitions and which specify the location where
    source code for the asset can be found.
    """

    code_references: Optional[CodeReferencesMetadataValue] = None

    @classmethod
    def namespace(cls) -> str:
        return "dagster"


def _with_code_source_single_definition(
    assets_def: Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"],
) -> Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"]:
    from dagster._core.definitions.assets import AssetsDefinition

    # SourceAsset and AssetSpec don't have an op definition to point to - cacheable assets
    # will be supported eventually but are a bit trickier
    if not isinstance(assets_def, AssetsDefinition):
        return assets_def

    metadata_by_key = dict(assets_def.metadata_by_key) or {}

    from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction
    from dagster._core.definitions.graph_definition import GraphDefinition
    from dagster._core.definitions.op_definition import OpDefinition

    base_fn = None
    if isinstance(assets_def.node_def, OpDefinition):
        base_fn = (
            assets_def.node_def.compute_fn.decorated_fn
            if isinstance(assets_def.node_def.compute_fn, DecoratedOpFunction)
            else assets_def.node_def.compute_fn
        )
    elif isinstance(assets_def.node_def, GraphDefinition):
        # For graph-backed assets, point to the composition fn, e.g. the
        # function decorated by @graph_asset
        base_fn = assets_def.node_def.composition_fn

    if not base_fn:
        return assets_def

    source_path = local_source_path_from_fn(base_fn)

    if source_path:
        sources = [source_path]

        for key in assets_def.keys:
            # merge with any existing metadata
            existing_source_code_metadata = CodeReferencesMetadataSet.extract(
                metadata_by_key.get(key, {})
            )
            existing_code_references = (
                existing_source_code_metadata.code_references.code_references
                if existing_source_code_metadata.code_references
                else []
            )
            sources_for_asset: list[Union[LocalFileCodeReference, UrlCodeReference]] = [
                *existing_code_references,
                *sources,
            ]

            metadata_by_key[key] = {
                **metadata_by_key.get(key, {}),
                **CodeReferencesMetadataSet(
                    code_references=CodeReferencesMetadataValue(code_references=sources_for_asset)
                ),
            }

    return assets_def.map_asset_specs(
        lambda spec: spec.replace_attributes(metadata=metadata_by_key[spec.key])
    )


@beta
class FilePathMapping(ABC):
    """Base class which defines a file path mapping function. These functions are used to map local file paths
    to their corresponding paths in a source control repository.

    In many cases where a source control repository is reproduced exactly on a local machine, the included
    AnchorBasedFilePathMapping class can be used to specify a direct mapping between the local file paths and the
    repository paths. However, in cases where the repository structure differs from the local structure, a custom
    mapping function can be provided to handle these cases.
    """

    @public
    @abstractmethod
    def convert_to_source_control_path(self, local_path: Path) -> str:
        """Maps a local file path to the corresponding path in a source control repository.

        Args:
            local_path (Path): The local file path to map.

        Returns:
            str: The corresponding path in the hosted source control repository, relative to the repository root.
        """


@beta
@dataclass
class AnchorBasedFilePathMapping(FilePathMapping):
    """Specifies the mapping between local file paths and their corresponding paths in a source control repository,
    using a specific file "anchor" as a reference point. All other paths are calculated relative to this anchor file.

    For example, if the chosen anchor file is `/Users/dagster/Documents/python_modules/my_module/my-module/__init__.py`
    locally, and `python_modules/my_module/my-module/__init__.py` in a source control repository, in order to map a
    different file `/Users/dagster/Documents/python_modules/my_module/my-module/my_asset.py` to the repository path,
    the mapping function will position the file in the repository relative to the anchor file's position in the repository,
    resulting in `python_modules/my_module/my-module/my_asset.py`.

    Args:
        local_file_anchor (Path): The path to a local file that is present in the repository.
        file_anchor_path_in_repository (str): The path to the anchor file in the repository.

    Example:
        .. code-block:: python

            mapping_fn = AnchorBasedFilePathMapping(
                local_file_anchor=Path(__file__),
                file_anchor_path_in_repository="python_modules/my_module/my-module/__init__.py",
            )
    """

    local_file_anchor: Path
    file_anchor_path_in_repository: str

    @public
    def convert_to_source_control_path(self, local_path: Path) -> str:
        """Maps a local file path to the corresponding path in a source control repository
        based on the anchor file and its corresponding path in the repository.

        Args:
            local_path (Path): The local file path to map.

        Returns:
            str: The corresponding path in the hosted source control repository, relative to the repository root.
        """
        path_from_anchor_to_target = os.path.relpath(
            local_path,
            self.local_file_anchor,
        )
        return os.path.normpath(
            os.path.join(
                self.file_anchor_path_in_repository,
                path_from_anchor_to_target,
            )
        )


def convert_local_path_to_git_path(
    base_git_url: str,
    file_path_mapping: FilePathMapping,
    local_path: LocalFileCodeReference,
) -> UrlCodeReference:
    source_file_from_repo_root = file_path_mapping.convert_to_source_control_path(
        Path(local_path.file_path)
    )
    line_number_suffix = f"#L{local_path.line_number}" if local_path.line_number else ""

    return UrlCodeReference(
        url=f"{base_git_url}/{source_file_from_repo_root}{line_number_suffix}",
        label=local_path.label,
    )


def _convert_local_path_to_git_path_single_definition(
    base_git_url: str,
    file_path_mapping: FilePathMapping,
    assets_def: Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"],
) -> Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"]:
    from dagster._core.definitions.assets import AssetsDefinition

    # SourceAsset doesn't have an op definition to point to - cacheable assets
    # will be supported eventually but are a bit trickier
    if not isinstance(assets_def, AssetsDefinition):
        return assets_def

    metadata_by_key = dict(assets_def.metadata_by_key) or {}

    for key in assets_def.keys:
        existing_source_code_metadata = CodeReferencesMetadataSet.extract(
            metadata_by_key.get(key, {})
        )
        if not existing_source_code_metadata.code_references:
            continue

        sources_for_asset: list[Union[LocalFileCodeReference, UrlCodeReference]] = [
            convert_local_path_to_git_path(
                base_git_url,
                file_path_mapping,
                source,
            )
            if isinstance(source, LocalFileCodeReference)
            else source
            for source in existing_source_code_metadata.code_references.code_references
        ]

        metadata_by_key[key] = {
            **metadata_by_key.get(key, {}),
            **CodeReferencesMetadataSet(
                code_references=CodeReferencesMetadataValue(code_references=sources_for_asset)
            ),
        }

    return assets_def.map_asset_specs(
        lambda spec: spec.replace_attributes(metadata=metadata_by_key[spec.key])
    )


def _build_github_url(url: str, branch: str) -> str:
    return f"{url}/tree/{branch}"


def _build_gitlab_url(url: str, branch: str) -> str:
    return f"{url}/-/tree/{branch}"


@beta
def link_code_references_to_git(
    assets_defs: Sequence[
        Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"]
    ],
    git_url: str,
    git_branch: str,
    file_path_mapping: FilePathMapping,
) -> Sequence[Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"]]:
    """Wrapper function which converts local file path code references to source control URLs
    based on the provided source control URL and branch.

    Args:
        assets_defs (Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition]]):
            The asset definitions to which source control metadata should be attached.
            Only assets with local file code references (such as those created by
            `with_source_code_references`) will be converted.
        git_url (str): The base URL for the source control system. For example,
            "https://github.com/dagster-io/dagster".
        git_branch (str): The branch in the source control system, such as "master".
        file_path_mapping (FilePathMapping):
            Specifies the mapping between local file paths and their corresponding paths in a source control repository.
            Simple usage is to provide a `AnchorBasedFilePathMapping` instance, which specifies an anchor file in the
            repository and the corresponding local file path, which is extrapolated to all other local file paths.
            Alternatively, a custom function can be provided which takes a local file path and returns the corresponding
            path in the repository, allowing for more complex mappings.

    Example:
        .. code-block:: python

                defs = Definitions(
                    assets=link_code_references_to_git(
                        with_source_code_references([my_dbt_assets]),
                        git_url="https://github.com/dagster-io/dagster",
                        git_branch="master",
                        file_path_mapping=AnchorBasedFilePathMapping(
                            local_file_anchor=Path(__file__),
                            file_anchor_path_in_repository="python_modules/my_module/my-module/__init__.py",
                        ),
                    )
                )
    """
    if "gitlab" in git_url:
        git_url = _build_gitlab_url(git_url, git_branch)
    elif "github.com" in git_url:
        git_url = _build_github_url(git_url, git_branch)
    else:
        raise ValueError(
            "Invalid `git_url`."
            " Only GitHub and GitLab are supported for linking to source control at this time."
        )

    return [
        _convert_local_path_to_git_path_single_definition(
            base_git_url=git_url,
            file_path_mapping=file_path_mapping,
            assets_def=assets_def,
        )
        for assets_def in assets_defs
    ]


@beta
def with_source_code_references(
    assets_defs: Sequence[
        Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"]
    ],
) -> Sequence[Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"]]:
    """Wrapper function which attaches local code reference metadata to the provided asset definitions.
    This points to the filepath and line number where the asset body is defined.

    Args:
        assets_defs (Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition]]):
            The asset definitions to which source code metadata should be attached.

    Returns:
        Sequence[AssetsDefinition]: The asset definitions with source code metadata attached.
    """
    return [_with_code_source_single_definition(assets_def) for assets_def in assets_defs]



import itertools
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, NamedTuple, Optional, Union, cast  # noqa: UP035

from dagster import _check as check
from dagster._core.definitions.composition import MappedInputPlaceholder
from dagster._core.definitions.dependency import (
    DependencyDefinition,
    DynamicCollectDependencyDefinition,
    GraphNode,
    IDependencyDefinition,
    MultiDependencyDefinition,
    NodeHandle,
    NodeInvocation,
    NodeOutput,
)
from dagster._core.definitions.graph_definition import GraphDefinition, SubselectedGraphDefinition
from dagster._core.errors import DagsterInvalidSubsetError
from dagster._core.selector.subset_selector import parse_op_queries

if TYPE_CHECKING:
    from dagster._core.definitions.node_definition import NodeDefinition


class OpSelection:
    def __init__(self, query: Iterable[str]):
        self.query = query

    def resolve(self, graph_def: GraphDefinition) -> AbstractSet[str]:
        if any(["." in item for item in self.query]):
            resolved_node_paths = set(self.query)
            _validate_node_paths(resolved_node_paths, graph_def)
        else:
            # validation happens inside parse_op_queries
            resolved_node_paths = set(parse_op_queries(graph_def, list(self.query)))
        return resolved_node_paths


class OpSelectionNode(NamedTuple):
    name: str
    children: list["OpSelectionNode"]

    @property
    def is_leaf(self) -> bool:
        return not self.children

    def has_child(self, name: str) -> bool:
        return any(child.name == name for child in self.children)

    def get_child(self, name: str) -> "OpSelectionNode":
        return next(child for child in self.children if child.name == name)


def _validate_node_paths(node_paths: Iterable[str], graph_def: GraphDefinition) -> None:
    selection_tree = _node_paths_to_tree(node_paths)
    _validate_selection_tree(selection_tree, graph_def)


def _node_paths_to_tree(node_paths: Iterable[str]) -> OpSelectionNode:
    node_path_lists = sorted([tuple(path.split(".")) for path in node_paths])
    return _node_path_lists_to_tree(node_path_lists)


def _node_path_lists_to_tree(paths: Sequence[tuple[str, ...]]) -> OpSelectionNode:
    root = OpSelectionNode("ROOT", [])
    for k, group in itertools.groupby(paths, lambda x: x[0]):
        child = _node_path_lists_to_tree([x[1:] for x in group if len(x) > 1])._replace(name=k)
        root.children.append(child)
    return root


def _validate_selection_tree(selection_tree: OpSelectionNode, graph_def: GraphDefinition) -> None:
    for selection_child in selection_tree.children:
        node_name = selection_child.name
        if not graph_def.has_node_named(node_name):
            raise DagsterInvalidSubsetError(
                f"Node {node_name} was selected, but no node named {node_name} was found."
            )
        if not selection_child.is_leaf:
            graph_child = graph_def.node_named(node_name)
            if isinstance(graph_child, GraphNode):
                _validate_selection_tree(selection_child, graph_child.definition)
            else:
                raise DagsterInvalidSubsetError(
                    f"Children of node {node_name} were selected, but {node_name} is not a graph."
                )


def get_graph_subset(
    graph: GraphDefinition,
    op_selection: Iterable[str],
    selected_outputs_by_op_handle: Mapping[NodeHandle, AbstractSet[str]],
) -> SubselectedGraphDefinition:
    """Returns a subsetted version of the given graph.

    Args:
        selected_outputs_by_op_handle: A mapping of op handles to the set of output names that are
            selected within that op. If an op handle isn't included in this dict, it means that
            _all_ its outputs are selected.
    """
    node_paths = OpSelection(op_selection).resolve(graph)
    selection_tree = _node_paths_to_tree(node_paths)
    return _get_graph_subset(
        graph,
        selection_tree,
        parent_handle=None,
        selected_outputs_by_op_handle=selected_outputs_by_op_handle,
    )


def _get_graph_subset(
    graph: GraphDefinition,
    selection_tree: OpSelectionNode,
    parent_handle: Optional[NodeHandle],
    selected_outputs_by_op_handle: Mapping[NodeHandle, AbstractSet[str]],
) -> SubselectedGraphDefinition:
    subgraph_deps: dict[
        NodeInvocation,
        dict[str, IDependencyDefinition],
    ] = {}

    subgraph_nodes: dict[str, NodeDefinition] = {}

    for node in graph.nodes_in_topological_order:
        # skip if the node isn't selected
        if not selection_tree.has_child(node.name):
            continue

        node_handle = NodeHandle(node.name, parent=parent_handle)
        node_def: Union[SubselectedGraphDefinition, NodeDefinition] = node.definition
        node_selection_tree = selection_tree.get_child(node.name)

        # subselect graph if any nodes inside the graph are selected
        if isinstance(node, GraphNode) and not node_selection_tree.is_leaf:
            node_def = _get_graph_subset(
                node.definition,
                node_selection_tree,
                parent_handle=node_handle,
                selected_outputs_by_op_handle=selected_outputs_by_op_handle,
            )

        subgraph_nodes[node.name] = node_def

        def is_output_selected(node_output: NodeOutput) -> bool:
            if not selection_tree.has_child(node_output.node_name):
                return False

            output_def, op_handle = graph.node_named(
                node_output.node_name
            ).definition.resolve_output_to_origin(
                node_output.output_name, NodeHandle(node_output.node_name, parent_handle)
            )
            return (
                op_handle not in selected_outputs_by_op_handle
                or output_def.name in selected_outputs_by_op_handle[check.not_none(op_handle)]
            )

        # build dependencies for the node. we do it for both cases because nested graphs can have
        # inputs and outputs too
        node_deps: dict[str, IDependencyDefinition] = {}
        for node_input in node.inputs():
            if graph.dependency_structure.has_direct_dep(node_input):
                node_output = graph.dependency_structure.get_direct_dep(node_input)
                if is_output_selected(node_output):
                    node_deps[node_input.input_name] = DependencyDefinition(
                        node=node_output.node.name, output=node_output.output_name
                    )
            elif graph.dependency_structure.has_dynamic_fan_in_dep(node_input):
                node_output = graph.dependency_structure.get_dynamic_fan_in_dep(node_input)
                if is_output_selected(node_output):
                    node_deps[node_input.input_name] = DynamicCollectDependencyDefinition(
                        node_name=node_output.node_name,
                        output_name=node_output.output_name,
                    )
            elif graph.dependency_structure.has_fan_in_deps(node_input):
                outputs = graph.dependency_structure.get_fan_in_deps(node_input)
                multi_dependencies = [
                    DependencyDefinition(
                        node=output_handle.node.name, output=output_handle.output_def.name
                    )
                    for output_handle in outputs
                    if (isinstance(output_handle, NodeOutput) and is_output_selected(output_handle))
                ]
                node_deps[node_input.input_name] = MultiDependencyDefinition(
                    cast(
                        list[Union[DependencyDefinition, type[MappedInputPlaceholder]]],
                        multi_dependencies,
                    )
                )
            # else input is unconnected

        dep_key = NodeInvocation(
            name=node.definition.name,
            alias=node.name,
            tags=node.tags,
            hook_defs=node.hook_defs,
            retry_policy=node.retry_policy,
        )
        subgraph_deps[dep_key] = node_deps

    # filter out unselected input/output mapping
    subgraph_input_mappings = [
        imap for imap in graph.input_mappings if imap.maps_to.node_name in subgraph_nodes.keys()
    ]
    subgraph_output_mappings = [
        omap for omap in graph.output_mappings if omap.maps_from.node_name in subgraph_nodes.keys()
    ]

    return SubselectedGraphDefinition(
        parent_graph_def=graph,
        dependencies=subgraph_deps,
        node_defs=list(subgraph_nodes.values()),
        input_mappings=subgraph_input_mappings,
        output_mappings=subgraph_output_mappings,
    )



from collections.abc import Mapping, Sequence, Set
from enum import Enum
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, Optional, TypeVar

from dagster._core.definitions.events import AssetKey
from dagster._record import record
from dagster._serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.definitions.remote_asset_graph import RemoteAssetGraph


@whitelist_for_serdes
class AssetDefinitionChangeType(Enum):
    """What change an asset has undergone between two deployments. Used
    in distinguishing asset definition changes in branch deployment and
    in subsequent other deployments.
    """

    NEW = "NEW"
    CODE_VERSION = "CODE_VERSION"
    DEPENDENCIES = "DEPENDENCIES"
    PARTITIONS_DEFINITION = "PARTITIONS_DEFINITION"
    TAGS = "TAGS"
    METADATA = "METADATA"
    REMOVED = "REMOVED"


class MappedKeyDiff(NamedTuple):
    added: Set[str]
    changed: Set[str]
    removed: Set[str]


T = TypeVar("T")


@whitelist_for_serdes
@record
class ValueDiff(Generic[T]):
    old: T
    new: T


@whitelist_for_serdes
@record
class DictDiff(Generic[T]):
    added_keys: Set[T]
    changed_keys: Set[T]
    removed_keys: Set[T]


@whitelist_for_serdes
@record
class AssetDefinitionDiffDetails:
    """Represents the diff information for changes between assets.

    Change types in change_types should have diff info for their corresponding fields
    if diff info is requested.

    "NEW" and "REMOVED" change types do not have diff info.
    """

    change_types: Set[AssetDefinitionChangeType]
    code_version: Optional[ValueDiff[Optional[str]]] = None
    dependencies: Optional[DictDiff[AssetKey]] = None
    partitions_definition: Optional[ValueDiff[Optional[str]]] = None
    tags: Optional[DictDiff[str]] = None
    metadata: Optional[DictDiff[str]] = None


class AssetGraphDiffer:
    """Given two asset graphs, base_asset_graph and branch_asset_graph, we can compute how the
    assets in branch_asset_graph have changed with respect to base_asset_graph. The ChangeReason
    enum contains the list of potential changes an asset can undergo. If the base_asset_graph is None,
    this indicates that the branch_asset_graph does not yet exist in the base deployment. In this case
    we will consider every asset New.
    """

    def __init__(
        self,
        branch_asset_graph: "RemoteAssetGraph",
        base_asset_graph: "RemoteAssetGraph",
    ):
        self._branch_asset_graph = branch_asset_graph
        self._base_asset_graph = base_asset_graph

    def _compare_base_and_branch_assets(
        self, asset_key: "AssetKey", include_diff: bool = False
    ) -> AssetDefinitionDiffDetails:
        """Computes the diff between a branch deployment asset and the
        corresponding base deployment asset.
        """
        if not self._base_asset_graph.has(asset_key):
            # if the base asset graph is None, it is because the asset graph in the branch deployment
            # is new and doesn't exist in the base deployment. Thus all assets are new.
            return AssetDefinitionDiffDetails(change_types={AssetDefinitionChangeType.NEW})

        if not self._branch_asset_graph.has(asset_key):
            return AssetDefinitionDiffDetails(change_types={AssetDefinitionChangeType.REMOVED})

        base_asset = self._base_asset_graph.get(asset_key).resolve_to_singular_repo_scoped_node()

        # This may not detect the case where an asset is moved from one repository or code location
        # to another as a change.
        branch_asset = self._branch_asset_graph.get(
            asset_key
        ).resolve_to_singular_repo_scoped_node()

        change_types: set[AssetDefinitionChangeType] = set()
        code_version_diff: Optional[ValueDiff] = None
        dependencies_diff: Optional[DictDiff] = None
        partitions_definition_diff: Optional[ValueDiff] = None
        tags_diff: Optional[DictDiff] = None
        metadata_diff: Optional[DictDiff] = None

        if branch_asset.code_version != base_asset.code_version:
            change_types.add(AssetDefinitionChangeType.CODE_VERSION)
            if include_diff:
                code_version_diff = ValueDiff(
                    old=base_asset.code_version, new=branch_asset.code_version
                )

        if branch_asset.parent_keys != base_asset.parent_keys:
            change_types.add(AssetDefinitionChangeType.DEPENDENCIES)
            if include_diff:
                dependencies_diff = DictDiff(
                    added_keys=branch_asset.parent_keys - base_asset.parent_keys,
                    changed_keys=set(),
                    removed_keys=base_asset.parent_keys - branch_asset.parent_keys,
                )

        else:
            # if the set of upstream dependencies is different, then we don't need to check if the partition mappings
            # for dependencies have changed since ChangeReason.DEPENDENCIES is already in the list of changes
            for upstream_asset_key in branch_asset.parent_keys:
                if branch_asset.partition_mappings.get(
                    upstream_asset_key
                ) != base_asset.partition_mappings.get(upstream_asset_key):
                    change_types.add(AssetDefinitionChangeType.DEPENDENCIES)
                    if include_diff:
                        dependencies_diff = DictDiff(
                            added_keys=set(),
                            changed_keys={upstream_asset_key},
                            removed_keys=set(),
                        )
                    break

        if branch_asset.partitions_def != base_asset.partitions_def:
            change_types.add(AssetDefinitionChangeType.PARTITIONS_DEFINITION)
            if include_diff:
                partitions_definition_diff = ValueDiff(
                    old=type(base_asset.partitions_def).__name__
                    if base_asset.partitions_def
                    else None,
                    new=type(branch_asset.partitions_def).__name__
                    if branch_asset.partitions_def
                    else None,
                )

        if branch_asset.tags != base_asset.tags:
            change_types.add(AssetDefinitionChangeType.TAGS)
            if include_diff:
                added, changed, removed = self._get_map_keys_diff(
                    base_asset.tags, branch_asset.tags
                )
                tags_diff = DictDiff(added_keys=added, changed_keys=changed, removed_keys=removed)

        if branch_asset.metadata != base_asset.metadata:
            change_types.add(AssetDefinitionChangeType.METADATA)
            if include_diff:
                added, changed, removed = self._get_map_keys_diff(
                    base_asset.metadata, branch_asset.metadata
                )
                metadata_diff = DictDiff(
                    added_keys=added, changed_keys=changed, removed_keys=removed
                )

        return AssetDefinitionDiffDetails(
            change_types=change_types,
            code_version=code_version_diff,
            dependencies=dependencies_diff,
            partitions_definition=partitions_definition_diff,
            tags=tags_diff,
            metadata=metadata_diff,
        )

    def _get_map_keys_diff(
        self, base: Mapping[str, Any], branch: Mapping[str, Any]
    ) -> MappedKeyDiff:
        """Returns the added, changed, and removed keys in the branch map compared to the base map."""
        added = set(branch.keys()) - set(base.keys())
        changed = {key for key in branch.keys() if key in base and branch[key] != base[key]}
        removed = set(base.keys()) - set(branch.keys())
        return MappedKeyDiff(added, changed, removed)

    def get_changes_for_asset(self, asset_key: "AssetKey") -> Sequence[AssetDefinitionChangeType]:
        """Returns list of AssetDefinitionChangeType for asset_key as compared to the base deployment."""
        return list(self._compare_base_and_branch_assets(asset_key).change_types)

    def get_changes_for_asset_with_diff(self, asset_key: "AssetKey") -> AssetDefinitionDiffDetails:
        """Returns list of AssetDefinitionDiff for asset_key as compared to the base deployment."""
        return self._compare_base_and_branch_assets(asset_key, include_diff=True)
